import unittest

import os
import sys
import shutil
import time

import grpc

from protoplasm import errors

import logging
log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

HERE = os.path.dirname(__file__)

PROTO_ROOT = os.path.join(HERE, 'res', 'proto')
BUILD_ROOT = os.path.join(HERE, 'res', 'build')


class ServiceTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Remove old stuff...
        build_package = os.path.join(BUILD_ROOT, 'unittesting')
        if os.path.exists(build_package):
            shutil.rmtree(build_package)
            time.sleep(0.1)

        from neobuilder.neobuilder import NeoBuilder

        # Build stuff...
        builder = NeoBuilder(package='unittesting',
                             protopath=PROTO_ROOT,
                             build_root=BUILD_ROOT)
        builder.build()

        # Add build root to path to access its modules
        sys.path.append(BUILD_ROOT)

    def test_raw_server_and_raw_client(self):
        port = '50042'
        log.info(f'Beginning...')
        from tests.servicetestutils.raw_server import UnaryServiceServer
        server = UnaryServiceServer()
        server.start(f'[::]:{port}')

        from tests.servicetestutils import raw_client
        self.assertEqual(server.servicer_implementation.calls['WithNoData'], 0)
        raw_client.call_WithNoData(f'localhost:{port}')
        self.assertEqual(server.servicer_implementation.calls['WithNoData'], 1)
        raw_client.call_WithNoData(f'localhost:{port}')
        self.assertEqual(server.servicer_implementation.calls['WithNoData'], 2)

        self.assertEqual(server.servicer_implementation.calls['WithInput'], 0)
        raw_client.call_WithInput(f'localhost:{port}', unnamed_input='did I win?')
        self.assertEqual(server.servicer_implementation.calls['WithInput'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithOutput'], 0)
        res = raw_client.call_WithOutput(f'localhost:{port}')
        self.assertEqual(res, 'you win')
        self.assertEqual(server.servicer_implementation.calls['WithOutput'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithBoth'], 0)
        res = raw_client.call_WithBoth(f'localhost:{port}', some_input='reverse me please')
        self.assertEqual(res, 'esaelp em esrever')
        self.assertEqual(server.servicer_implementation.calls['WithBoth'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithManyInputs'], 0)
        raw_client.call_WithManyInputs(f'localhost:{port}', first_input='yay', second_input=42, third_input=True)
        self.assertEqual(server.servicer_implementation.calls['WithManyInputs'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithManyOutputs'], 0)
        res_a, res_b, res_c = raw_client.call_WithManyOutputs(f'localhost:{port}')
        self.assertEqual(res_a, 'snorlax')
        self.assertEqual(res_b, 7)
        self.assertTrue(res_c)
        self.assertEqual(server.servicer_implementation.calls['WithManyOutputs'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithManyBoths'], 0)
        res_a, res_b, res_c = raw_client.call_WithManyBoths(f'localhost:{port}',
                                                            another_first_input='bar',
                                                            another_second_input=42,
                                                            another_third_input=True)
        self.assertEqual(res_a, 'BAR')
        self.assertEqual(res_b, 21)
        self.assertFalse(res_c)
        self.assertEqual(server.servicer_implementation.calls['WithManyBoths'], 1)

        # Now lets break it...
        self.assertEqual(server.servicer_implementation.calls['WithInput'], 1)
        with self.assertRaises(grpc.RpcError) as cm:
            raw_client.call_WithInput(f'localhost:{port}', unnamed_input='explode')
        self.assertEqual(cm.exception.code(), grpc.StatusCode.NOT_FOUND)  # noqa
        self.assertEqual(cm.exception.details(), 'totally fake not found error')  # noqa
        self.assertEqual(server.servicer_implementation.calls['WithInput'], 2)

        raw_client.call_WithInput(f'localhost:{port}', unnamed_input='nevermind')
        self.assertEqual(server.servicer_implementation.calls['WithInput'], 3)

        server.stop()

    def test_raw_server_and_protoplasm_client(self):
        port = '50043'
        log.info(f'Beginning...')
        from tests.servicetestutils.raw_server import UnaryServiceServer
        server = UnaryServiceServer()
        server.start(f'[::]:{port}')

        from unittesting.unary.unaryservice_grpc_sender import UnaryService
        client = UnaryService(f'localhost:{port}')

        self.assertEqual(server.servicer_implementation.calls['WithNoData'], 0)
        client.with_no_data()
        self.assertEqual(server.servicer_implementation.calls['WithNoData'], 1)
        client.with_no_data()
        self.assertEqual(server.servicer_implementation.calls['WithNoData'], 2)

        self.assertEqual(server.servicer_implementation.calls['WithInput'], 0)
        client.with_input('did I win?')
        self.assertEqual(server.servicer_implementation.calls['WithInput'], 1)
        with self.assertRaises(errors.NotFound) as cm:
            client.with_input('explode')
        self.assertEqual(cm.exception.status_code, grpc.StatusCode.NOT_FOUND)
        self.assertEqual(cm.exception.details, 'totally fake not found error')
        self.assertEqual(server.servicer_implementation.calls['WithInput'], 2)

        self.assertEqual(server.servicer_implementation.calls['WithOutput'], 0)
        res = client.with_output()
        self.assertEqual(res, 'you win')
        self.assertEqual(server.servicer_implementation.calls['WithOutput'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithBoth'], 0)
        res = client.with_both('reverse me please')
        self.assertEqual(res, 'esaelp em esrever')
        self.assertEqual(server.servicer_implementation.calls['WithBoth'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithManyInputs'], 0)
        client.with_many_inputs(first_input='yay', second_input=42, third_input=True)
        self.assertEqual(server.servicer_implementation.calls['WithManyInputs'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithManyOutputs'], 0)
        res_a, res_b, res_c = client.with_many_outputs()
        self.assertEqual(res_a, 'snorlax')
        self.assertEqual(res_b, 7)
        self.assertTrue(res_c)
        self.assertEqual(server.servicer_implementation.calls['WithManyOutputs'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithManyBoths'], 0)
        res_a, res_b, res_c = client.with_many_boths(another_first_input='bar',
                                                     another_second_input=42,
                                                     another_third_input=True)
        self.assertEqual(res_a, 'BAR')
        self.assertEqual(res_b, 21)
        self.assertFalse(res_c)
        self.assertEqual(server.servicer_implementation.calls['WithManyBoths'], 1)

        server.stop()

    def test_protoplasm_server_and_raw_client(self):
        port = '50041'
        log.info(f'Beginning...')
        from tests.servicetestutils.plasm_server import UnaryProtoplasmServer
        server = UnaryProtoplasmServer()
        server.start(f'[::]:{port}')

        from tests.servicetestutils import raw_client
        self.assertEqual(server.servicer_implementation.calls['WithNoData'], 0)
        raw_client.call_WithNoData(f'localhost:{port}')
        self.assertEqual(server.servicer_implementation.calls['WithNoData'], 1)
        raw_client.call_WithNoData(f'localhost:{port}')
        self.assertEqual(server.servicer_implementation.calls['WithNoData'], 2)

        self.assertEqual(server.servicer_implementation.calls['WithInput'], 0)
        raw_client.call_WithInput(f'localhost:{port}', unnamed_input='did I win?')
        self.assertEqual(server.servicer_implementation.calls['WithInput'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithOutput'], 0)
        res = raw_client.call_WithOutput(f'localhost:{port}')
        self.assertEqual(res, 'you win')
        self.assertEqual(server.servicer_implementation.calls['WithOutput'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithBoth'], 0)
        res = raw_client.call_WithBoth(f'localhost:{port}', some_input='reverse me please')
        self.assertEqual(res, 'esaelp em esrever')
        self.assertEqual(server.servicer_implementation.calls['WithBoth'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithManyInputs'], 0)
        raw_client.call_WithManyInputs(f'localhost:{port}', first_input='yay', second_input=42, third_input=True)
        self.assertEqual(server.servicer_implementation.calls['WithManyInputs'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithManyOutputs'], 0)
        res_a, res_b, res_c = raw_client.call_WithManyOutputs(f'localhost:{port}')
        self.assertEqual(res_a, 'snorlax')
        self.assertEqual(res_b, 7)
        self.assertTrue(res_c)
        self.assertEqual(server.servicer_implementation.calls['WithManyOutputs'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithManyBoths'], 0)
        res_a, res_b, res_c = raw_client.call_WithManyBoths(f'localhost:{port}',
                                                            another_first_input='bar',
                                                            another_second_input=42,
                                                            another_third_input=True)
        self.assertEqual(res_a, 'BAR')
        self.assertEqual(res_b, 21)
        self.assertFalse(res_c)
        self.assertEqual(server.servicer_implementation.calls['WithManyBoths'], 1)

        # Now lets break it...
        self.assertEqual(server.servicer_implementation.calls['WithInput'], 1)
        with self.assertRaises(grpc.RpcError) as cm:
            raw_client.call_WithInput(f'localhost:{port}', unnamed_input='explode')
        self.assertEqual(cm.exception.code(), grpc.StatusCode.NOT_FOUND)  # noqa
        self.assertEqual(cm.exception.details(), 'totally fake not found error')  # noqa
        self.assertEqual(server.servicer_implementation.calls['WithInput'], 2)

        raw_client.call_WithInput(f'localhost:{port}', unnamed_input='nevermind')
        self.assertEqual(server.servicer_implementation.calls['WithInput'], 3)

        server.stop()

    def test_protoplasm_server_and_protoplasm_client(self):
        port = '50044'
        log.info(f'Beginning...')
        from tests.servicetestutils.plasm_server import UnaryProtoplasmServer
        server = UnaryProtoplasmServer()
        server.start(f'[::]:{port}')

        from unittesting.unary.unaryservice_grpc_sender import UnaryService
        client = UnaryService(f'localhost:{port}')

        self.assertEqual(server.servicer_implementation.calls['WithNoData'], 0)
        client.with_no_data()
        self.assertEqual(server.servicer_implementation.calls['WithNoData'], 1)
        client.with_no_data()
        self.assertEqual(server.servicer_implementation.calls['WithNoData'], 2)

        self.assertEqual(server.servicer_implementation.calls['WithInput'], 0)
        client.with_input('did I win?')
        self.assertEqual(server.servicer_implementation.calls['WithInput'], 1)
        with self.assertRaises(errors.NotFound) as cm:
            client.with_input('explode')
        self.assertEqual(cm.exception.status_code, grpc.StatusCode.NOT_FOUND)
        self.assertEqual(cm.exception.details, 'totally fake not found error')
        self.assertEqual(server.servicer_implementation.calls['WithInput'], 2)

        self.assertEqual(server.servicer_implementation.calls['WithOutput'], 0)
        res = client.with_output()
        self.assertEqual(res, 'you win')
        self.assertEqual(server.servicer_implementation.calls['WithOutput'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithBoth'], 0)
        res = client.with_both('reverse me please')
        self.assertEqual(res, 'esaelp em esrever')
        self.assertEqual(server.servicer_implementation.calls['WithBoth'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithManyInputs'], 0)
        client.with_many_inputs(first_input='yay', second_input=42, third_input=True)
        self.assertEqual(server.servicer_implementation.calls['WithManyInputs'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithManyOutputs'], 0)
        res_a, res_b, res_c = client.with_many_outputs()
        self.assertEqual(res_a, 'snorlax')
        self.assertEqual(res_b, 7)
        self.assertTrue(res_c)
        self.assertEqual(server.servicer_implementation.calls['WithManyOutputs'], 1)

        self.assertEqual(server.servicer_implementation.calls['WithManyBoths'], 0)
        res_a, res_b, res_c = client.with_many_boths(another_first_input='bar',
                                                     another_second_input=42,
                                                     another_third_input=True)
        self.assertEqual(res_a, 'BAR')
        self.assertEqual(res_b, 21)
        self.assertFalse(res_c)
        self.assertEqual(server.servicer_implementation.calls['WithManyBoths'], 1)

        server.stop()



__all__ = [
    'GrpcExceptionWrapper',
]

import grpc
from typing import *
from protoplasm import errors

_STATUS_CODE_TO_EXCEPTION_MAP = {
    grpc.StatusCode.UNKNOWN: errors.api.Unknown,
    grpc.StatusCode.NOT_FOUND: errors.api.NotFound,
    grpc.StatusCode.INVALID_ARGUMENT: errors.api.InvalidArgument,
    grpc.StatusCode.UNIMPLEMENTED: errors.api.Unimplemented,
    grpc.StatusCode.PERMISSION_DENIED: errors.api.PermissionDenied,
    grpc.StatusCode.UNAUTHENTICATED: errors.api.Unauthenticated,
    grpc.StatusCode.UNAVAILABLE: errors.api.Unavailable,
    grpc.StatusCode.DEADLINE_EXCEEDED: errors.api.DeadlineExceeded,
    grpc.StatusCode.ALREADY_EXISTS: errors.api.AlreadyExists,
    grpc.StatusCode.FAILED_PRECONDITION: errors.api.FailedPrecondition,
    grpc.StatusCode.DATA_LOSS: errors.api.DataLoss,
    grpc.StatusCode.RESOURCE_EXHAUSTED: errors.api.ResourceExhausted,
    grpc.StatusCode.OUT_OF_RANGE: errors.api.OutOfRange,
    grpc.StatusCode.INTERNAL: errors.api.Internal,
    grpc.StatusCode.CANCELLED: errors.api.Cancelled,
    grpc.StatusCode.ABORTED: errors.api.Aborted,
}


class GrpcExceptionWrapper(object):
    def __init__(self, ex: grpc.RpcError):
        self.ex = ex
        self._code = False
        self._debug_error_string = False
        self._details = False
        self._initial_metadata = False
        self._traceback = False
        self._trailing_metadata = False

    @property
    def code(self) -> Optional[grpc.StatusCode]:
        if self._code is False:
            self._code = None
            code_call = getattr(self.ex, 'code', None)
            if code_call and hasattr(code_call, '__call__'):
                self._code = code_call()
        return self._code

    @property
    def debug_error_string(self) -> str:
        if self._debug_error_string is False:
            self._debug_error_string = ''
            code_call = getattr(self.ex, 'debug_error_string', None)
            if code_call and hasattr(code_call, '__call__'):
                self._debug_error_string = code_call()
        return self._debug_error_string

    @property
    def details(self) -> str:
        if self._details is False:
            self._details = ''
            code_call = getattr(self.ex, 'details', None)
            if code_call and hasattr(code_call, '__call__'):
                self._details = code_call()
        return self._details

    @property
    def initial_metadata(self) -> dict:
        if self._initial_metadata is False:
            self._initial_metadata = {}
            code_call = getattr(self.ex, 'initial_metadata', None)
            if code_call and hasattr(code_call, '__call__'):
                for k, v in code_call():
                    k = k.lower()
                    if k not in self._initial_metadata:
                        self._initial_metadata[k] = v
                    else:
                        if not isinstance(self._initial_metadata[k], list):
                            self._initial_metadata[k] = [self._initial_metadata[k]]
                        self._initial_metadata[k].append(v)
        return self._initial_metadata

    @property
    def traceback(self):
        if self._traceback is False:
            self._traceback = None
            code_call = getattr(self.ex, 'traceback', None)
            if code_call and hasattr(code_call, '__call__'):
                self._traceback = code_call()
        return self._traceback

    @property
    def trailing_metadata(self) -> dict:
        if self._trailing_metadata is False:
            self._trailing_metadata = {}
            code_call = getattr(self.ex, 'trailing_metadata', None)
            if code_call and hasattr(code_call, '__call__'):
                for k, v in code_call():
                    k = k.lower()
                    if k not in self._trailing_metadata:
                        self._trailing_metadata[k] = v
                    else:
                        if not isinstance(self._trailing_metadata[k], list):
                            self._trailing_metadata[k] = [self._trailing_metadata[k]]
                        self._trailing_metadata[k].append(v)
        return self._trailing_metadata

    def get_reraise(self):
        return _STATUS_CODE_TO_EXCEPTION_MAP.get(self.code, errors.api.Unknown)(self.details,
                                                                                ref_code=self.trailing_metadata.get('ref_code', None))







import base64
import datetime
import unittest
import os
import sys

from protoplasm import casting

import logging
log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

import shutil
import time
HERE = os.path.dirname(__file__)
PROTO_ROOT = os.path.join(HERE, 'res', 'proto')
BUILD_ROOT = os.path.join(HERE, 'res', 'build')


class CastingTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    @classmethod
    def setUpClass(cls) -> None:
        # Remove old stuff...
        build_package = os.path.join(BUILD_ROOT, 'sandbox')
        if os.path.exists(build_package):
            shutil.rmtree(build_package)
            time.sleep(0.1)

        from neobuilder.neobuilder import NeoBuilder

        # Build stuff...
        builder = NeoBuilder(package='sandbox',
                             protopath=PROTO_ROOT,
                             build_root=BUILD_ROOT)
        builder.build()

        # Add build root to path to access its modules
        sys.path.append(BUILD_ROOT)

    def test_proto_to_dataclass_and_back(self):
        from sandbox.test import rainbow_pb2
        from sandbox.test import rainbow_dc

        p1 = rainbow_pb2.SubMessage()
        p1.foo = 'Foo!'
        p1.bar = 'Bar!!!'

        dc1 = rainbow_dc.SubMessage()
        dc1.foo = 'Foo!'
        dc1.bar = 'Bar!!!'

        self.assertEqual(p1, casting.dataclass_to_proto(dc1))
        self.assertEqual(dc1, casting.proto_to_dataclass(p1))

        p2 = rainbow_pb2.SubMessage()
        p2.foo = 'Foo Two!'

        dc2 = rainbow_dc.SubMessage()
        dc2.foo = 'Foo Two!'

        self.assertEqual(p2, casting.dataclass_to_proto(dc2))
        self.assertEqual(dc2, casting.proto_to_dataclass(p2))

        p3 = rainbow_pb2.SubMessage()
        p3.foo = '这是中国人'

        dc3 = rainbow_dc.SubMessage()
        dc3.foo = '这是中国人'

        self.assertEqual(p3, casting.dataclass_to_proto(dc3))
        self.assertEqual(dc3, casting.proto_to_dataclass(p3))

        p4 = rainbow_pb2.SubMessage()
        dc4 = rainbow_dc.SubMessage()

        self.assertEqual(p4, casting.dataclass_to_proto(dc4))
        self.assertEqual(dc4, casting.proto_to_dataclass(p4))

        self.assertNotEqual(dc1, casting.proto_to_dataclass(p2))

        self.assertEqual(dc1, casting.proto_to_dataclass(casting.dataclass_to_proto(dc1)))
        self.assertEqual(p1, casting.dataclass_to_proto(casting.proto_to_dataclass(p1)))

    def test_nested_proto_to_dataclass_and_back(self):
        from sandbox.test import rainbow_pb2
        from sandbox.test import rainbow_dc

        p1 = rainbow_pb2.RainbowMessage()
        dc1 = rainbow_dc.RainbowMessage()

        self.assertEqual(p1, casting.dataclass_to_proto(dc1))
        self.assertEqual(dc1, casting.proto_to_dataclass(p1))
        self.assertEqual(dc1, casting.proto_to_dataclass(casting.dataclass_to_proto(dc1)))
        self.assertEqual(p1, casting.dataclass_to_proto(casting.proto_to_dataclass(p1)))

        p4 = rainbow_pb2.RainbowMessage()
        p4.simple_field = 'Green'
        p4.message_field.foo = 'Blue'
        p4.message_field.bar = 'Yellow'
        
        dc4 = rainbow_dc.RainbowMessage()
        dc4.simple_field = 'Green'
        dc4.message_field = rainbow_dc.SubMessage()
        dc4.message_field.foo = 'Blue'
        dc4.message_field.bar = 'Yellow'

        self.assertEqual(p4, casting.dataclass_to_proto(dc4))
        self.assertEqual(dc4, casting.proto_to_dataclass(p4))
        self.assertEqual(dc4, casting.proto_to_dataclass(casting.dataclass_to_proto(dc4)))
        self.assertEqual(p4, casting.dataclass_to_proto(casting.proto_to_dataclass(p4)))
        
        p10 = rainbow_pb2.RainbowMessage()
        p10.simple_field = 'Green'
        p10.message_field.foo = 'Blue'
        p10.message_field.bar = 'Yellow'
        p10.simple_list.append('Four')
        p10.simple_list.append('Seven')
        p10.simple_list.append('99')
        m7 = p10.message_list.add()
        m7.foo = 'Foo in a list'
        m7.bar = 'Candybar'
        m8 = p10.message_list.add()
        m8.foo = 'a Fool in a list'
        m8.bar = 'Tequila bar'
        p10.simple_map['dora'] = 'Imamap!'
        p10.simple_map['diego'] = 'Camera!'
        p10.message_map['mickey'].foo = 'mouse'
        p10.message_map['donald'].foo = 'duck'
        p10.message_map['donald'].bar = 'trump'

        dc10 = rainbow_dc.RainbowMessage()
        dc10.simple_field = 'Green'
        dc10.message_field = rainbow_dc.SubMessage(foo='Blue', bar='Yellow')
        dc10.simple_list.append('Four')
        dc10.simple_list.append('Seven')
        dc10.simple_list.append('99')
        dc10.message_list.append(rainbow_dc.SubMessage(foo='Foo in a list', bar='Candybar'))
        dc10.message_list.append(rainbow_dc.SubMessage(foo='a Fool in a list', bar='Tequila bar'))
        dc10.simple_map['dora'] = 'Imamap!'
        dc10.simple_map['diego'] = 'Camera!'
        dc10.message_map['mickey'] = rainbow_dc.SubMessage(foo='mouse')
        dc10.message_map['donald'] = rainbow_dc.SubMessage(foo='duck', bar='trump')

        self.assertEqual(p10, casting.dataclass_to_proto(dc10))
        self.assertEqual(dc10, casting.proto_to_dataclass(p10))
        self.assertEqual(dc10, casting.proto_to_dataclass(casting.dataclass_to_proto(dc10)))
        self.assertEqual(p10, casting.dataclass_to_proto(casting.proto_to_dataclass(p10)))

    def test_timestamp_proto_to_dataclass_and_back(self):
        from sandbox.test import rainbow_pb2
        from sandbox.test import rainbow_dc

        p1 = rainbow_pb2.TimestampMessage()
        dc1 = rainbow_dc.TimestampMessage()

        self.assertEqual(p1, casting.dataclass_to_proto(dc1))
        self.assertEqual(dc1, casting.proto_to_dataclass(p1))
        self.assertEqual(dc1, casting.proto_to_dataclass(casting.dataclass_to_proto(dc1)))
        self.assertEqual(p1, casting.dataclass_to_proto(casting.proto_to_dataclass(p1)))

        ts1 = '2012-07-03T14:50:51.654321Z'
        dt1 = datetime.datetime(2012, 7, 3, 14, 50, 51, 654321)
        self.assertEqual(ts1, dt1.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts2 = '2010-04-09T22:38:39.009870Z'
        dt2 = datetime.datetime(2010, 4, 9, 22, 38, 39, 9870)
        self.assertEqual(ts2, dt2.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts3 = '1979-07-06T14:30:00.000000Z'
        dt3 = datetime.datetime(1979, 7, 6, 14, 30)
        self.assertEqual(ts3, dt3.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts4 = '2049-01-01T12:00:00.000000Z'
        dt4 = datetime.datetime(2049, 1, 1, 12)
        self.assertEqual(ts4, dt4.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts5 = '2049-01-01T00:00:00.000000Z'
        dt5 = datetime.datetime(2049, 1, 1)
        self.assertEqual(ts5, dt5.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        p_expect = rainbow_pb2.TimestampMessage()
        p_expect.my_timestamp.FromJsonString(ts1)
        p_expect.my_timestamp_list.add().FromJsonString(ts2)
        p_expect.my_timestamp_list.add().FromJsonString(ts3)
        p_expect.my_timestamp_map['noon'].FromJsonString(ts4)
        p_expect.my_timestamp_map['midnight'].FromJsonString(ts5)

        dc_expect = rainbow_dc.TimestampMessage()
        dc_expect.my_timestamp = dt1
        dc_expect.my_timestamp_list.append(dt2)
        dc_expect.my_timestamp_list.append(dt3)
        dc_expect.my_timestamp_map['noon'] = dt4
        dc_expect.my_timestamp_map['midnight'] = dt5

        self.assertEqual(p_expect, casting.dataclass_to_proto(dc_expect))
        self.assertEqual(dc_expect, casting.proto_to_dataclass(p_expect))
        self.assertEqual(dc_expect, casting.proto_to_dataclass(casting.dataclass_to_proto(dc_expect)))
        self.assertEqual(p_expect, casting.dataclass_to_proto(casting.proto_to_dataclass(p_expect)))

    def test_byte_proto_to_dataclass_and_back(self):
        from sandbox.test import rainbow_pb2
        from sandbox.test import rainbow_dc

        p1 = rainbow_pb2.BytesMessage()
        dc1 = rainbow_dc.BytesMessage()

        self.assertEqual(p1, casting.dataclass_to_proto(dc1))
        self.assertEqual(dc1, casting.proto_to_dataclass(p1))
        self.assertEqual(dc1, casting.proto_to_dataclass(casting.dataclass_to_proto(dc1)))
        self.assertEqual(p1, casting.dataclass_to_proto(casting.proto_to_dataclass(p1)))

        as_str_1 = 'Þórður Matthíasson'
        as_bytes_1 = as_str_1.encode('utf-8')
        as_base64_1 = base64.encodebytes(as_bytes_1).decode('utf-8').strip()

        as_str_2 = '♡'
        as_bytes_2 = as_str_2.encode('utf-8')
        as_base64_2 = base64.encodebytes(as_bytes_2).decode('utf-8').strip()

        as_str_3 = '♠'
        as_bytes_3 = as_str_3.encode('utf-8')
        as_base64_3 = base64.encodebytes(as_bytes_3).decode('utf-8').strip()

        as_str_4 = '♢'
        as_bytes_4 = as_str_4.encode('utf-8')
        as_base64_4 = base64.encodebytes(as_bytes_4).decode('utf-8').strip()

        as_str_5 = '♣'
        as_bytes_5 = as_str_5.encode('utf-8')
        as_base64_5 = base64.encodebytes(as_bytes_5).decode('utf-8').strip()

        p_expect = rainbow_pb2.BytesMessage()
        p_expect.my_bytes = as_bytes_1
        p_expect.my_bytes_list.append(as_bytes_2)
        p_expect.my_bytes_list.append(as_bytes_3)
        p_expect.my_bytes_map['zero'] = as_bytes_4
        p_expect.my_bytes_map['one'] = as_bytes_5

        dc_expect = rainbow_dc.BytesMessage()
        dc_expect.my_bytes = as_bytes_1
        dc_expect.my_bytes_list.append(as_bytes_2)
        dc_expect.my_bytes_list.append(as_bytes_3)
        dc_expect.my_bytes_map['zero'] = as_bytes_4
        dc_expect.my_bytes_map['one'] = as_bytes_5

        self.assertEqual(p_expect, casting.dataclass_to_proto(dc_expect))
        self.assertEqual(dc_expect, casting.proto_to_dataclass(p_expect))
        self.assertEqual(dc_expect, casting.proto_to_dataclass(casting.dataclass_to_proto(dc_expect)))
        self.assertEqual(p_expect, casting.dataclass_to_proto(casting.proto_to_dataclass(p_expect)))

    def test_enum_proto_to_dataclass_and_back(self):
        from sandbox.test import enums_pb2
        from sandbox.test import enums_dc

        p1 = enums_pb2.WithExternalEnum()
        dc1 = enums_dc.WithExternalEnum()

        self.assertEqual(p1, casting.dataclass_to_proto(dc1))
        self.assertEqual(dc1, casting.proto_to_dataclass(p1))
        self.assertEqual(dc1, casting.proto_to_dataclass(casting.dataclass_to_proto(dc1)))
        self.assertEqual(p1, casting.dataclass_to_proto(casting.proto_to_dataclass(p1)))

        p_expect = enums_pb2.WithExternalEnum()
        p_expect.my_enum = enums_pb2.TWO

        p_expect.my_enum_list.append(enums_pb2.ONE)
        p_expect.my_enum_list.append(enums_pb2.THREE)

        p_expect.my_enum_map['one'] = enums_pb2.ONE
        p_expect.my_enum_map['two'] = enums_pb2.TWO

        p_expect.my_alias_enum = enums_pb2.SIX

        p_expect.my_alias_enum_list.append(enums_pb2.FOUR)
        p_expect.my_alias_enum_list.append(enums_pb2.FIVE)
        p_expect.my_alias_enum_list.append(enums_pb2.FJORIR)

        p_expect.my_alias_enum_map['six'] = enums_pb2.SIX
        p_expect.my_alias_enum_map['sex'] = enums_pb2.SEX
        p_expect.my_alias_enum_map['fimm'] = enums_pb2.FIVE

        dc_expect = enums_dc.WithExternalEnum()
        dc_expect.my_enum = enums_dc.TWO

        dc_expect.my_enum_list.append(enums_dc.ONE)
        dc_expect.my_enum_list.append(enums_dc.THREE)

        dc_expect.my_enum_map['one'] = enums_dc.ONE
        dc_expect.my_enum_map['two'] = enums_dc.TWO

        dc_expect.my_alias_enum = enums_dc.SIX

        dc_expect.my_alias_enum_list.append(enums_dc.FOUR)
        dc_expect.my_alias_enum_list.append(enums_dc.FIVE)
        dc_expect.my_alias_enum_list.append(enums_dc.FJORIR)

        dc_expect.my_alias_enum_map['six'] = enums_dc.SIX
        dc_expect.my_alias_enum_map['sex'] = enums_dc.SEX
        dc_expect.my_alias_enum_map['fimm'] = enums_dc.FIVE

        self.assertEqual(p_expect, casting.dataclass_to_proto(dc_expect))
        self.assertEqual(dc_expect, casting.proto_to_dataclass(p_expect))
        self.assertEqual(dc_expect, casting.proto_to_dataclass(casting.dataclass_to_proto(dc_expect)))
        self.assertEqual(p_expect, casting.dataclass_to_proto(casting.proto_to_dataclass(p_expect)))

        p_expect2 = enums_pb2.WithInternalEnum()
        p_expect2.my_internal_enum = enums_pb2.WithInternalEnum.SIX

        p_expect2.my_internal_enum_list.append(enums_pb2.WithInternalEnum.FIVE)
        p_expect2.my_internal_enum_list.append(enums_pb2.WithInternalEnum.FOUR)

        p_expect2.my_internal_enum_map['no5'] = enums_pb2.WithInternalEnum.FIVE
        p_expect2.my_internal_enum_map['no6'] = enums_pb2.WithInternalEnum.SIX

        p_expect2.my_internal_alias_enum = enums_pb2.WithInternalEnum.SEVEN

        p_expect2.my_internal_alias_enum_list.append(enums_pb2.WithInternalEnum.SJO)
        p_expect2.my_internal_alias_enum_list.append(enums_pb2.WithInternalEnum.ATTA)
        p_expect2.my_internal_alias_enum_list.append(enums_pb2.WithInternalEnum.EIGHT)

        p_expect2.my_internal_alias_enum_map['no9'] = enums_pb2.WithInternalEnum.NIU
        p_expect2.my_internal_alias_enum_map['no9B'] = enums_pb2.WithInternalEnum.NINE
        p_expect2.my_internal_alias_enum_map['default'] = enums_pb2.WithInternalEnum.ZERO

        dc_expect2 = enums_dc.WithInternalEnum()
        dc_expect2.my_internal_enum = enums_dc.WithInternalEnum.SIX

        dc_expect2.my_internal_enum_list.append(enums_dc.WithInternalEnum.FIVE)
        dc_expect2.my_internal_enum_list.append(enums_dc.WithInternalEnum.FOUR)

        dc_expect2.my_internal_enum_map['no5'] = enums_dc.WithInternalEnum.FIVE
        dc_expect2.my_internal_enum_map['no6'] = enums_dc.WithInternalEnum.SIX

        dc_expect2.my_internal_alias_enum = enums_dc.WithInternalEnum.SEVEN

        dc_expect2.my_internal_alias_enum_list.append(enums_dc.WithInternalEnum.SJO)
        dc_expect2.my_internal_alias_enum_list.append(enums_dc.WithInternalEnum.ATTA)
        dc_expect2.my_internal_alias_enum_list.append(enums_dc.WithInternalEnum.EIGHT)

        dc_expect2.my_internal_alias_enum_map['no9'] = enums_dc.WithInternalEnum.NIU
        dc_expect2.my_internal_alias_enum_map['no9B'] = enums_dc.WithInternalEnum.NINE
        dc_expect2.my_internal_alias_enum_map['default'] = enums_dc.WithInternalEnum.ZERO

        self.assertEqual(p_expect2, casting.dataclass_to_proto(dc_expect2))
        self.assertEqual(dc_expect2, casting.proto_to_dataclass(p_expect2))
        self.assertEqual(dc_expect2, casting.proto_to_dataclass(casting.dataclass_to_proto(dc_expect2)))
        self.assertEqual(p_expect2, casting.dataclass_to_proto(casting.proto_to_dataclass(p_expect2)))

    def test_duration_proto_to_dataclass_and_back(self):
        from sandbox.test import timeduration_pb2
        from sandbox.test import timeduration_dc

        p1 = timeduration_pb2.DurationMessage()
        dc1 = timeduration_dc.DurationMessage()

        self.assertEqual(p1, casting.dataclass_to_proto(dc1))
        self.assertEqual(dc1, casting.proto_to_dataclass(p1))
        self.assertEqual(dc1, casting.proto_to_dataclass(casting.dataclass_to_proto(dc1)))
        self.assertEqual(p1, casting.dataclass_to_proto(casting.proto_to_dataclass(p1)))

        p_expect = timeduration_pb2.DurationMessage()
        p_expect.my_duration.seconds = 987
        p_expect.my_duration.nanos = 654321000

        dl_1 = p_expect.my_duration_list.add()
        dl_1.seconds = 0
        dl_1.nanos = 42000

        dl_2 = p_expect.my_duration_list.add()
        dl_2.seconds = 100001

        p_expect.my_duration_map['short'].nanos = 7000
        p_expect.my_duration_map['long'].seconds = 60*60*24*4  # 4 days

        dc_expect = timeduration_dc.DurationMessage(datetime.timedelta(seconds=987.654321),
                                                    [
                                                        datetime.timedelta(seconds=0.000042),
                                                        datetime.timedelta(seconds=100001)
                                                    ],
                                                    {
                                                        'short': datetime.timedelta(microseconds=7),
                                                        'long': datetime.timedelta(days=4)
                                                    })

        self.assertEqual(p_expect, casting.dataclass_to_proto(dc_expect))
        self.assertEqual(dc_expect, casting.proto_to_dataclass(p_expect))
        self.assertEqual(dc_expect, casting.proto_to_dataclass(casting.dataclass_to_proto(dc_expect)))
        self.assertEqual(p_expect, casting.dataclass_to_proto(casting.proto_to_dataclass(p_expect)))

    def test_any_proto_to_dataclass_and_back(self):
        from sandbox.test import rainbow_pb2
        from sandbox.test import rainbow_dc
        from sandbox.test import anytest_pb2
        from sandbox.test import anytest_dc

        p1 = anytest_pb2.AnyMessage()
        dc1 = anytest_dc.AnyMessage()

        self.assertEqual(p1, casting.dataclass_to_proto(dc1))
        self.assertEqual(dc1, casting.proto_to_dataclass(p1))
        self.assertEqual(dc1, casting.proto_to_dataclass(casting.dataclass_to_proto(dc1)))
        self.assertEqual(p1, casting.dataclass_to_proto(casting.proto_to_dataclass(p1)))

        p2_sub = rainbow_pb2.SubMessage()
        p2_sub.foo = 'OOF'
        p2_sub.bar = 'RAB'

        p2_expect = anytest_pb2.AnyMessage()
        p2_expect.my_any.Pack(p2_sub)

        dc2_sub = rainbow_dc.SubMessage('OOF', 'RAB')
        dc2 = anytest_dc.AnyMessage()
        dc2.my_any = dc2_sub

        self.assertEqual(p2_sub, casting.dataclass_to_proto(dc2_sub))
        self.assertEqual(dc2_sub, casting.proto_to_dataclass(p2_sub))
        self.assertEqual(dc2_sub, casting.proto_to_dataclass(casting.dataclass_to_proto(dc2_sub)))
        self.assertEqual(p2_sub, casting.dataclass_to_proto(casting.proto_to_dataclass(p2_sub)))

        self.assertEqual(p2_expect, casting.dataclass_to_proto(dc2))
        self.assertEqual(dc2, casting.proto_to_dataclass(p2_expect))
        self.assertEqual(dc2, casting.proto_to_dataclass(casting.dataclass_to_proto(dc2)))
        self.assertEqual(p2_expect, casting.dataclass_to_proto(casting.proto_to_dataclass(p2_expect)))

        i1 = p2_expect.my_any_list.add()
        i1.Pack(casting.mkproto(rainbow_pb2.SubMessage, foo='too', bar='beer'))
        i2 = p2_expect.my_any_list.add()
        i2.Pack(casting.mkproto(rainbow_pb2.SubMessage, foo='three', bar='tequila'))

        self.assertNotEqual(p2_expect, casting.dataclass_to_proto(dc2))
        self.assertNotEqual(dc2, casting.proto_to_dataclass(p2_expect))

        dc2.my_any_list = [rainbow_dc.SubMessage(foo='too', bar='beer'), rainbow_dc.SubMessage(foo='three', bar='tequila')]

        self.assertEqual(p2_expect, casting.dataclass_to_proto(dc2))
        self.assertEqual(dc2, casting.proto_to_dataclass(p2_expect))
        self.assertEqual(dc2, casting.proto_to_dataclass(casting.dataclass_to_proto(dc2)))
        self.assertEqual(p2_expect, casting.dataclass_to_proto(casting.proto_to_dataclass(p2_expect)))

        p2_expect.my_any_map['mars'].Pack(casting.mkproto(rainbow_pb2.SubMessage, foo='xy'))
        p2_expect.my_any_map['venus'].Pack(casting.mkproto(rainbow_pb2.SubMessage, foo='xx'))

        self.assertNotEqual(p2_expect, casting.dataclass_to_proto(dc2))
        self.assertNotEqual(dc2, casting.proto_to_dataclass(p2_expect))

        dc2.my_any_map['mars'] = rainbow_dc.SubMessage('xy')
        dc2.my_any_map['venus'] = rainbow_dc.SubMessage('xx')

        self.assertEqual(p2_expect, casting.dataclass_to_proto(dc2))
        self.assertEqual(dc2, casting.proto_to_dataclass(p2_expect))
        self.assertEqual(dc2, casting.proto_to_dataclass(casting.dataclass_to_proto(dc2)))
        self.assertEqual(p2_expect, casting.dataclass_to_proto(casting.proto_to_dataclass(p2_expect)))

    def test_any_nested_proto_to_dataclass_and_back(self):
        from sandbox.test import rainbow_pb2
        from sandbox.test import rainbow_dc
        from sandbox.test import anytest_pb2
        from sandbox.test import anytest_dc

        p1 = anytest_pb2.AnyMessage()
        dc1 = anytest_dc.AnyMessage()

        self.assertEqual(p1, casting.dataclass_to_proto(dc1))
        self.assertEqual(dc1, casting.proto_to_dataclass(p1))
        self.assertEqual(dc1, casting.proto_to_dataclass(casting.dataclass_to_proto(dc1)))
        self.assertEqual(p1, casting.dataclass_to_proto(casting.proto_to_dataclass(p1)))

        p10 = rainbow_pb2.RainbowMessage()
        p10.simple_field = 'Green'
        p10.message_field.foo = 'Blue'
        p10.message_field.bar = 'Yellow'
        p10.simple_list.append('Four')
        p10.simple_list.append('Seven')
        p10.simple_list.append('99')
        m7 = p10.message_list.add()
        m7.foo = 'Foo in a list'
        m7.bar = 'Candybar'
        m8 = p10.message_list.add()
        m8.foo = 'a Fool in a list'
        m8.bar = 'Tequila bar'
        p10.simple_map['dora'] = 'Imamap!'
        p10.simple_map['diego'] = 'Camera!'
        p10.message_map['mickey'].foo = 'mouse'
        p10.message_map['donald'].foo = 'duck'
        p10.message_map['donald'].bar = 'trump'

        dc10 = rainbow_dc.RainbowMessage()
        dc10.simple_field = 'Green'
        dc10.message_field = rainbow_dc.SubMessage(foo='Blue', bar='Yellow')
        dc10.simple_list.append('Four')
        dc10.simple_list.append('Seven')
        dc10.simple_list.append('99')
        dc10.message_list.append(rainbow_dc.SubMessage(foo='Foo in a list', bar='Candybar'))
        dc10.message_list.append(rainbow_dc.SubMessage(foo='a Fool in a list', bar='Tequila bar'))
        dc10.simple_map['dora'] = 'Imamap!'
        dc10.simple_map['diego'] = 'Camera!'
        dc10.message_map['mickey'] = rainbow_dc.SubMessage(foo='mouse')
        dc10.message_map['donald'] = rainbow_dc.SubMessage(foo='duck', bar='trump')

        self.assertEqual(p10, casting.dataclass_to_proto(dc10))
        self.assertEqual(dc10, casting.proto_to_dataclass(p10))
        self.assertEqual(dc10, casting.proto_to_dataclass(casting.dataclass_to_proto(dc10)))
        self.assertEqual(p10, casting.dataclass_to_proto(casting.proto_to_dataclass(p10)))

        p2_expect = anytest_pb2.AnyMessage()
        p2_expect.my_any.Pack(p10)

        dc2 = anytest_dc.AnyMessage()
        dc2.my_any = dc10

        self.assertEqual(p2_expect, casting.dataclass_to_proto(dc2))
        self.assertEqual(dc2, casting.proto_to_dataclass(p2_expect))
        self.assertEqual(dc2, casting.proto_to_dataclass(casting.dataclass_to_proto(dc2)))

        # TODO(thordurm@ccpgames.com>) 2024-04-15: These tests are derpy in Debian...
        #  seems like the binary serializer there orders the dict fields differently and thus the serialized bytes are different!
        # log.info(f'p2_expect={p2_expect!r}')
        # log.info(f'p2_expect as dataclass={casting.proto_to_dataclass(p2_expect)!r}')
        # log.info(f'p2_expect and back again ={casting.dataclass_to_proto(casting.proto_to_dataclass(p2_expect))!r}')
        # log.info(f'p2_expect and back again as dataclass={casting.proto_to_dataclass(casting.dataclass_to_proto(casting.proto_to_dataclass(p2_expect)))!r}')
        #
        # self.assertEqual(p2_expect, casting.dataclass_to_proto(casting.proto_to_dataclass(p2_expect)))
        #
        # i1 = p2_expect.my_any_list.add()
        # i1.Pack(p10)
        # i2 = p2_expect.my_any_list.add()
        # i2.Pack(p10)
        #
        # self.assertNotEqual(p2_expect, casting.dataclass_to_proto(dc2))
        # self.assertNotEqual(dc2, casting.proto_to_dataclass(p2_expect))
        #
        # dc2.my_any_list = [dc10, dc10]
        #
        # self.assertEqual(p2_expect, casting.dataclass_to_proto(dc2))
        # self.assertEqual(dc2, casting.proto_to_dataclass(p2_expect))
        # self.assertEqual(dc2, casting.proto_to_dataclass(casting.dataclass_to_proto(dc2)))
        # self.assertEqual(p2_expect, casting.dataclass_to_proto(casting.proto_to_dataclass(p2_expect)))
        #
        # p2_expect.my_any_map['mars'].Pack(p10)
        # p2_expect.my_any_map['venus'].Pack(p10)
        #
        # self.assertNotEqual(p2_expect, casting.dataclass_to_proto(dc2))
        # self.assertNotEqual(dc2, casting.proto_to_dataclass(p2_expect))
        #
        # dc2.my_any_map['mars'] = dc10
        # dc2.my_any_map['venus'] = dc10
        #
        # self.assertEqual(p2_expect, casting.dataclass_to_proto(dc2))
        # self.assertEqual(dc2, casting.proto_to_dataclass(p2_expect))
        # self.assertEqual(dc2, casting.proto_to_dataclass(casting.dataclass_to_proto(dc2)))
        # self.assertEqual(p2_expect, casting.dataclass_to_proto(casting.proto_to_dataclass(p2_expect)))

    def test_cloning(self):
        from sandbox.test import clones_dc

        thing_1 = clones_dc.ThingOne(
            my_string='this is a string',
            my_number=42,
            my_float=4.2,
            my_timestamp=datetime.datetime(2019, 3, 28, 10, 13, 8, 123456),
            my_subthing=clones_dc.SubThing(foo='F', bar='B'),
            my_unique_string='I REFUSE TO BE CLONED!!!'
        )

        thing_2 = clones_dc.ThingTwo.from_clone(thing_1)

        self.assertEqual(thing_1.my_string, thing_2.my_string)
        self.assertEqual(thing_1.my_number, thing_2.my_number)
        self.assertEqual(thing_1.my_float, thing_2.my_float)
        self.assertEqual(thing_1.my_timestamp, thing_2.my_timestamp)
        self.assertEqual(thing_1.my_subthing, thing_2.my_subthing)

        self.assertNotEqual(thing_1.my_unique_string, thing_2.my_special_string)



import unittest
import datetime
import dataclasses
import base64
import enum

import protoplasm.bases.dataclass_bases
from protoplasm.casting import dictators
from protoplasm import bases
import collections

import logging
log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

import os
import sys
import sys
import shutil
import time
HERE = os.path.dirname(__file__)
PROTO_ROOT = os.path.join(HERE, 'res', 'proto')
BUILD_ROOT = os.path.join(HERE, 'res', 'build')


class FauxEnum(enum.IntEnum):
    DEFAULT = 0
    FOO = 1
    BAR = 2
    OOF = 3
    RAB = 4


@dataclasses.dataclass
class FauxDataClass(protoplasm.bases.dataclass_bases.DataclassBase):
    string: str = dataclasses.field(default=None)
    integer: int = dataclasses.field(default=None)
    floating: float = dataclasses.field(default=None)
    boolean: bool = dataclasses.field(default=None)
    real_bytes: bytes = dataclasses.field(default=None)
    timestamp: datetime.datetime = dataclasses.field(default=None)
    enumerator: FauxEnum = dataclasses.field(default=None)
    duration: datetime.timedelta = dataclasses.field(default=None)
    any: protoplasm.bases.dataclass_bases.DataclassBase = dataclasses.field(default=None)


def _get_field_map(dc):
    return {f.name: f for f in dataclasses.fields(dc)}


def _get_test_args(dc, field_name):
    return getattr(dc, field_name), _get_field_map(dc)[field_name], dc


class BaseDictatorTest(unittest.TestCase):
    def test_none_to_dict_value(self):
        caster = dictators.BaseDictator.to_dict_value
        dc = FauxDataClass()
        self.assertIsNone(dc.string)
        self.assertIsNone(caster(*_get_test_args(dc, 'string')))

        self.assertIsNone(dc.integer)
        self.assertIsNone(caster(*_get_test_args(dc, 'integer')))

        self.assertIsNone(dc.floating)
        self.assertIsNone(caster(*_get_test_args(dc, 'floating')))

        self.assertIsNone(dc.boolean)
        self.assertIsNone(caster(*_get_test_args(dc, 'boolean')))

    def test_to_dict_value(self):
        caster = dictators.BaseDictator.to_dict_value
        dc = FauxDataClass(string='Ðiss is a júníkód stíng viþþ tjænís skvígglí bitts 中國潦草的位',
                           integer=4815162342,
                           floating=3.14159265359,
                           boolean=True)

        self.assertIsNotNone(dc.string)
        self.assertIsInstance(dc.string, str)
        self.assertEqual('Ðiss is a júníkód stíng viþþ tjænís skvígglí bitts 中國潦草的位',
                         caster(*_get_test_args(dc, 'string')))

        self.assertIsNotNone(dc.integer)
        self.assertIsInstance(dc.integer, int)
        self.assertEqual(4815162342, caster(*_get_test_args(dc, 'integer')))

        self.assertIsNotNone(dc.floating)
        self.assertIsInstance(dc.floating, float)
        self.assertEqual(3.14159265359, caster(*_get_test_args(dc, 'floating')))

        self.assertIsNotNone(dc.boolean)
        self.assertIsInstance(dc.boolean, bool)
        self.assertEqual(True, caster(*_get_test_args(dc, 'boolean')))

    def test_none_from_dict_value(self):
        caster = dictators.BaseDictator.from_dict_value
        dc = FauxDataClass()

        self.assertIsNone(dc.string)
        dc.string = caster(None, _get_field_map(dc)['string'], dc.__class__)
        self.assertIsNone(dc.string)

        self.assertIsNone(dc.integer)
        dc.integer = caster(None, _get_field_map(dc)['integer'], dc.__class__)
        self.assertIsNone(dc.integer)

        self.assertIsNone(dc.floating)
        dc.floating = caster(None, _get_field_map(dc)['floating'], dc.__class__)
        self.assertIsNone(dc.floating)

        self.assertIsNone(dc.boolean)
        dc.boolean = caster(None, _get_field_map(dc)['boolean'], dc.__class__)
        self.assertIsNone(dc.boolean)

    def test_from_dict_value(self):
        caster = dictators.BaseDictator.from_dict_value
        string = 'Ðiss is a júníkód stíng viþþ tjænís skvígglí bitts 中國潦草的位'
        integer = 4815162342
        floating = 3.14159265359
        boolean = True

        dc = FauxDataClass()

        self.assertIsNone(dc.string)
        dc.string = caster(string, _get_field_map(dc)['string'], dc)
        self.assertIsNotNone(dc.string)
        self.assertIsInstance(dc.string, str)
        self.assertEqual('Ðiss is a júníkód stíng viþþ tjænís skvígglí bitts 中國潦草的位', dc.string)

        self.assertIsNone(dc.integer)
        dc.integer = caster(integer, _get_field_map(dc)['integer'], dc)
        self.assertIsNotNone(dc.integer)
        self.assertIsInstance(dc.integer, int)
        self.assertEqual(4815162342, dc.integer)

        self.assertIsNone(dc.floating)
        dc.floating = caster(floating, _get_field_map(dc)['floating'], dc)
        self.assertIsNotNone(dc.floating)
        self.assertIsInstance(dc.floating, float)
        self.assertEqual(3.14159265359, dc.floating)

        self.assertIsNone(dc.boolean)
        dc.boolean = caster(boolean, _get_field_map(dc)['boolean'], dc)
        self.assertIsNotNone(dc.boolean)
        self.assertIsInstance(dc.boolean, bool)
        self.assertEqual(True, dc.boolean)


class TimestampDictatorTest(unittest.TestCase):
    def test_none_to_dict_value(self):
        caster = dictators.TimestampDictator.to_dict_value
        dc = FauxDataClass()
        self.assertIsNone(dc.timestamp)
        self.assertIsNone(caster(*_get_test_args(dc, 'timestamp')))

    def test_to_dict_value(self):
        caster = dictators.TimestampDictator.to_dict_value
        dc = FauxDataClass(timestamp=datetime.datetime(2010, 4, 9, 22, 38, 39, 987654))

        self.assertIsNotNone(dc.timestamp)
        self.assertIsInstance(dc.timestamp, datetime.datetime)
        self.assertEqual('2010-04-09T22:38:39.987654Z', caster(*_get_test_args(dc, 'timestamp')))

    def test_none_from_dict_value(self):
        caster = dictators.TimestampDictator.from_dict_value
        dc = FauxDataClass()

        self.assertIsNone(dc.timestamp)
        dc.timestamp = caster(None, _get_field_map(dc)['timestamp'], dc.__class__)
        self.assertIsNone(dc.timestamp)

    def test_from_dict_value(self):
        caster = dictators.TimestampDictator.from_dict_value
        timestamp_string = '2012-07-03T14:50:51.654321Z'

        dc = FauxDataClass()

        self.assertIsNone(dc.timestamp)
        dc.timestamp = caster(timestamp_string, _get_field_map(dc)['timestamp'], dc.__class__)
        self.assertIsNotNone(dc.timestamp)
        self.assertIsInstance(dc.timestamp, datetime.datetime)
        self.assertEqual(datetime.datetime(2012, 7, 3, 14, 50, 51, 654321), dc.timestamp)


class ByteDictatorTest(unittest.TestCase):
    def test_none_to_dict_value(self):
        caster = dictators.ByteDictator.to_dict_value
        dc = FauxDataClass()
        self.assertIsNone(dc.real_bytes)
        # self.assertEqual(b'', dc.real_bytes)
        # self.assertIsNone(caster(*_get_test_args(dc, 'real_bytes')))
        self.assertEqual('', caster(*_get_test_args(dc, 'real_bytes')))

    def test_to_dict_value(self):
        caster = dictators.ByteDictator.to_dict_value
        as_str = '♡♠♢♣'
        as_bytes = as_str.encode('utf-8')
        as_base64 = base64.encodebytes(as_bytes).decode('utf-8').strip()

        dc = FauxDataClass(real_bytes=as_bytes)

        self.assertIsNotNone(dc.real_bytes)
        self.assertIsInstance(dc.real_bytes, bytes)
        self.assertEqual(as_base64, caster(*_get_test_args(dc, 'real_bytes')))

    def test_none_from_dict_value(self):
        caster = dictators.ByteDictator.from_dict_value
        dc = FauxDataClass()

        self.assertIsNone(dc.real_bytes)
        # self.assertEqual(b'', dc.real_bytes)
        dc.real_bytes = caster(None, _get_field_map(dc)['real_bytes'], dc.__class__)
        # self.assertIsNone(dc.real_bytes)
        self.assertEqual(b'', dc.real_bytes)

    def test_from_dict_value(self):
        caster = dictators.ByteDictator.from_dict_value
        as_str = '♡♠♢♣'
        as_bytes = as_str.encode('utf-8')
        as_base64 = base64.encodebytes(as_bytes).decode('utf-8').strip()

        dc = FauxDataClass()

        self.assertIsNone(dc.real_bytes)
        dc.real_bytes = caster(as_base64, _get_field_map(dc)['real_bytes'], dc.__class__)
        self.assertIsNotNone(dc.real_bytes)
        self.assertIsInstance(dc.real_bytes, bytes)
        self.assertEqual(as_bytes, dc.real_bytes)


class EnumDictatorTest(unittest.TestCase):
    def test_none_to_dict_value(self):
        caster = dictators.EnumDictator.to_dict_value
        dc = FauxDataClass()

        self.assertIsNone(dc.enumerator)
        # self.assertEqual(0, dc.enumerator)
        # self.assertIsNone(caster(*_get_test_args(dc, 'enumerator')))
        self.assertEqual(0, caster(*_get_test_args(dc, 'enumerator')))

    def test_to_dict_value(self):
        caster = dictators.EnumDictator.to_dict_value
        dc = FauxDataClass(enumerator=FauxEnum.FOO)

        self.assertIsNotNone(dc.enumerator)
        self.assertIsInstance(dc.enumerator, enum.IntEnum)
        self.assertEqual(FauxEnum.FOO.value, caster(*_get_test_args(dc, 'enumerator')))

    def test_none_from_dict_value(self):
        caster = dictators.EnumDictator.from_dict_value
        dc = FauxDataClass()

        self.assertIsNone(dc.enumerator)
        # self.assertEqual(0, dc.enumerator)
        dc.enumerator = caster(None, _get_field_map(dc)['enumerator'], dc.__class__)
        # self.assertIsNone(dc.enumerator)
        self.assertEqual(FauxEnum.DEFAULT, dc.enumerator)

    def test_from_dict_value(self):
        caster = dictators.EnumDictator.from_dict_value
        dc = FauxDataClass()

        self.assertIsNone(dc.enumerator)
        dc.enumerator = caster(2, _get_field_map(dc)['enumerator'], dc.__class__)
        self.assertIsNotNone(dc.enumerator)
        self.assertIsInstance(dc.enumerator, enum.IntEnum)
        self.assertEqual(FauxEnum.BAR, dc.enumerator)

        dc2 = FauxDataClass()

        self.assertIsNone(dc2.enumerator)
        dc2.enumerator = caster('OOF', _get_field_map(dc2)['enumerator'], dc2.__class__)
        self.assertIsNotNone(dc2.enumerator)
        self.assertIsInstance(dc2.enumerator, enum.IntEnum)
        self.assertEqual(FauxEnum.OOF, dc2.enumerator)


class DurationDictatorTest(unittest.TestCase):
    def test_none_to_dict_value(self):
        caster = dictators.DurationDictator.to_dict_value
        dc = FauxDataClass()
        self.assertIsNone(dc.duration)
        self.assertIsNone(caster(*_get_test_args(dc, 'duration')))

    def test_to_dict_value(self):
        caster = dictators.DurationDictator.to_dict_value
        dc = FauxDataClass(duration=datetime.timedelta(seconds=12.3456789))

        self.assertIsNotNone(dc.duration)
        self.assertIsInstance(dc.duration, datetime.timedelta)
        self.assertEqual('12.345679s', caster(*_get_test_args(dc, 'duration')))

    def test_none_from_dict_value(self):
        caster = dictators.DurationDictator.from_dict_value
        dc = FauxDataClass()

        self.assertIsNone(dc.duration)
        dc.duration = caster(None, _get_field_map(dc)['duration'], dc.__class__)
        self.assertIsNone(dc.duration)

    def test_from_dict_value(self):
        caster = dictators.DurationDictator.from_dict_value
        duration_string = '12.345679s'

        dc = FauxDataClass()

        self.assertIsNone(dc.duration)
        dc.duration = caster(duration_string, _get_field_map(dc)['duration'], dc.__class__)
        self.assertIsNotNone(dc.duration)
        self.assertIsInstance(dc.duration, datetime.timedelta)
        self.assertEqual(datetime.timedelta(seconds=12.3456789), dc.duration)

        self.assertEqual(datetime.timedelta(seconds=12.3456789), caster(12.3456789, _get_field_map(dc)['duration'], dc.__class__))
        self.assertEqual(datetime.timedelta(seconds=123456), caster(123456, _get_field_map(dc)['duration'], dc.__class__))
        self.assertEqual(datetime.timedelta(seconds=12.3456789), caster('12.3456789', _get_field_map(dc)['duration'], dc.__class__))


class AnyDictatorTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Remove old stuff...
        build_package = os.path.join(BUILD_ROOT, 'sandbox')
        if os.path.exists(build_package):
            shutil.rmtree(build_package)
            time.sleep(0.1)

        from neobuilder.neobuilder import NeoBuilder

        # Build stuff...
        builder = NeoBuilder(package='sandbox',
                             protopath=PROTO_ROOT,
                             build_root=BUILD_ROOT)
        builder.build()

        # Add build root to path to access its modules
        sys.path.append(BUILD_ROOT)

    def test_none_to_dict_value(self):
        caster = dictators.AnyDictator.to_dict_value
        dc = FauxDataClass()
        self.assertIsNone(dc.any)
        self.assertIsNone(caster(*_get_test_args(dc, 'any')))

    def test_to_dict_value(self):
        from sandbox.test import rainbow_dc

        caster = dictators.AnyDictator.to_dict_value
        dc = FauxDataClass(any=rainbow_dc.SubMessage('FOO!', 'BAR!!'))

        self.assertIsNotNone(dc.any)
        self.assertIsInstance(dc.any, protoplasm.bases.dataclass_bases.DataclassBase)
        self.assertIsInstance(dc.any, rainbow_dc.SubMessage)

        expected = collections.OrderedDict([
            ('@type', 'type.googleapis.com/sandbox.test.SubMessage'),
            ('foo', 'FOO!'),
            ('bar', 'BAR!!')
        ])

        self.assertEqual(expected, caster(*_get_test_args(dc, 'any')))

    def test_none_from_dict_value(self):
        caster = dictators.AnyDictator.from_dict_value
        dc = FauxDataClass()

        self.assertIsNone(dc.any)
        dc.duration = caster(None, _get_field_map(dc)['any'], dc.__class__)
        self.assertIsNone(dc.any)

    def test_from_dict_value(self):
        from sandbox.test import anytest_dc
        from sandbox.test import rainbow_dc

        caster = dictators.AnyDictator.from_dict_value

        input_value = collections.OrderedDict([
            ('@type', 'type.evetech.net/sandbox.test.SubMessage'),
            ('foo', 'FOO!'),
            ('bar', 'BAR!!')
        ])

        dc = anytest_dc.AnyMessage()

        self.assertIsNone(dc.my_any)

        dc.my_any = caster(input_value, _get_field_map(dc)['my_any'], dc.__class__)

        self.assertIsNotNone(dc.my_any)
        self.assertIsInstance(dc.my_any, protoplasm.bases.dataclass_bases.DataclassBase)
        self.assertIsInstance(dc.my_any, rainbow_dc.SubMessage)
        self.assertEqual(rainbow_dc.SubMessage('FOO!', 'BAR!!'), dc.my_any)



__all__ = [
    'ProtoplasmError',
    'ServiceApiException',
    'ResponseIterationTimoutError',
]


class ProtoplasmError(Exception):
    pass


class ServiceApiException(ProtoplasmError):
    def __init__(self, status_code=2, details='unknown', ref_code=None, should_log=False):
        """Base Error/Exception raised by service API implementations that can
        not return normally.
        """
        self.status_code = status_code
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class ResponseIterationTimoutError(ProtoplasmError, TimeoutError):
    pass



__all__ = [
    'RequestIterator',
]
from ._base import *
from .dataclassbase import T_DCB
import queue


class RequestIterator(Generic[T_DCB]):
    def __init__(self, initial_request: Optional[Union[T_DCB, List[T_DCB], Tuple[T_DCB]]] = None):
        self._queue = queue.SimpleQueue()
        if initial_request:
            if isinstance(initial_request, (list, tuple)):
                for ir in initial_request:
                    self.send(ir)
            else:
                self.send(initial_request)
        self._do_cancel: bool = False

    def close(self):
        self._do_cancel = True

    def send(self, request: T_DCB):
        self._queue.put(request)

    def __iter__(self):
        return self

    def __next__(self) -> T_DCB:
        if self._do_cancel:
            raise StopIteration()
        return self._queue.get(block=True)



__all__ = [
    'IResponseIterator',
    'ResponseIterator',
    'TimeoutResponseIterator',
]
import abc
from ._base import *
from protoplasm.errors import *
from .dataclassbase import T_DCB
import threading
import grpc  # noqa

import logging
log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from protoplasm.bases._remotewrap import RemoteMethodWrapper  # noqa


class IResponseIterator(Iterable, abc.ABC):
    @abc.abstractmethod
    def close(self):
        pass

    @abc.abstractmethod
    def next(self) -> Optional[T_DCB]:
        pass

    def __iter__(self):
        return self

    def __next__(self) -> T_DCB:
        n = self.next()
        if n is None:
            raise StopIteration()
        return n


class ResponseIterator(IResponseIterator, Generic[T_DCB]):
    __slots__ = ('_wrapper', '_response_iterator')

    def __init__(self, remote_wrapper: 'RemoteMethodWrapper'):
        """Super simple basic Response Iterator to wrap gRPC RPC stream responses.

        Can be used as an iterator directly in a loop or by calling `next()`
        directly to get the next iteration from the result stream.

        Note that `next()` will simply block until it receives a value to return
        from the stream (unless the stream has been closed in which case
        `next()` returns None).
        """
        self._wrapper = remote_wrapper
        self._response_iterator = iter(remote_wrapper.iterable_results)

    def close(self):
        """Cancels the RPC call mid-stream.
        """
        self._wrapper.close()

    def next(self) -> Optional[T_DCB]:
        if self._wrapper.is_closed:
            return None
        return next(self._response_iterator)

    def with_timeout(self, timeout_seconds: Optional[float] = None,
                     raise_timeout: Optional[bool] = None) -> 'TimeoutResponseIterator':
        return TimeoutResponseIterator(self, timeout_seconds, raise_timeout)

    def one_and_done(self, timeout_seconds: Optional[float] = 0,
                     raise_timeout: Optional[bool] = None) -> Optional[T_DCB]:
        """Waits for a single iteration from the stream to return and then closes the stream.

        Note the default timeout of 0, meaning this will wait forever by default!
        """
        return self.with_timeout(timeout_seconds, raise_timeout).one_and_done()


class TimeoutResponseIterator(IResponseIterator, Generic[T_DCB]):
    __slots__ = ('_response_iterator', '_timeout_seconds', '_timeout_thread', '_next',
                 '_exception', '_timed_out', '_raise_timeout')

    def __init__(self, response_iterator: ResponseIterator[T_DCB],
                 timeout_seconds: float = 60.0, raise_timeout: bool = True):
        """Response Iterator wrapper with builtin timeout (deadline)
        while waiting for next stream iteration.

        Can be used as an iterator directly in a loop or by calling `next()`
        directly to get the next iteration from the result stream.

        Note that `next()` will block until it receives a value to return
        from the stream OR it times out.

        If raise_timeout is True then a ResponseIterationTimoutError will be
        raised in case of a timeout but if not, then the iteration loop will
        simply end if this is used as such (StopIteration is raised) or the
        timed-out `next()` call returns None.
        """
        self._response_iterator: ResponseIterator = response_iterator
        self._timeout_seconds = timeout_seconds
        if self._timeout_seconds is None:
            self._timeout_seconds = 60.0
        self._timeout_thread: Optional[threading.Thread] = None
        self._next: Optional[T_DCB] = None
        self._exception: Optional[Exception] = None
        self._timed_out: bool = False
        self._raise_timeout = raise_timeout
        if self._raise_timeout is None:
            self._raise_timeout = True

    def next(self, timeout_seconds: Optional[float] = None, raise_timeout: Optional[bool] = None) -> Optional[T_DCB]:
        """The `timeout_seconds` and `raise_timeout` parameters are optional
        and override the wrappers default ones given when it was initialized.

        Returns None if `raise_timeout` is False.
        """
        self._next = None
        self._exception = None
        self._timed_out = False
        if timeout_seconds is None:
            timeout_seconds = self._timeout_seconds
        if raise_timeout is None:
            raise_timeout = self._raise_timeout
        if not timeout_seconds:  # Just wait forever! :D
            return self._response_iterator.next()

        self._timeout_thread = threading.Thread(target=self._wait_for_next)
        self._timeout_thread.start()
        self._timeout_thread.join(timeout_seconds)
        if self._timeout_thread.is_alive():
            log.info('thread is still alive... closing!')
            self._timed_out = True
            self.close()
            if raise_timeout:
                raise ResponseIterationTimoutError(f'timout while waiting for next response from'
                                                   f' {self._response_iterator._wrapper.method_name}')  # noqa
            else:
                log.info(f'stopping iteration after timeout')
                return None

        if self._exception:  # Something bad happened!
            raise self._exception

        return self._next  # noqa

    def _wait_for_next(self):
        log.info('_wait_for_next started...')
        try:
            self._next = self._response_iterator.next()
            log.info('_next happened!!')
        except grpc.RpcError as ex:
            from protoplasm.bases._exwrap import GrpcExceptionWrapper  # noqa
            wrex = GrpcExceptionWrapper(ex)
            if wrex.code == grpc.StatusCode.CANCELLED and self._timed_out:
                log.info(f'Self inflicted timeout in thread')
                return  # this means we timed out and cancelled this ourselves!
            log.exception(f'RpcError in thread: {ex!r}')
            self._exception = wrex.get_reraise()
        except Exception as ex:
            log.exception(f'Exception in thread: {ex!r}')
            self._exception = ex
        log.info('_wait_for_next ended!!')

    def close(self):
        self._response_iterator.close()

    def one_and_done(self, timeout_seconds: Optional[float] = None,
                     raise_timeout: Optional[bool] = None) -> Optional[T_DCB]:
        """Waits for a single iteration from the stream to return and then closes the stream.
        """
        the_one = self.next(timeout_seconds, raise_timeout)
        self.close()
        return the_one



__all__ = [
    'dict_to_proto',
    'dict_to_dataclass',
]
from google.protobuf import json_format
from protoplasm.casting import dictators
from protoplasm.casting import castutils
from protoplasm.structs import *
import dataclasses

import logging

log = logging.getLogger(__name__)


def dict_to_proto(proto_class: Type[GeneratedProtocolMessageType],
                  dict_data: Optional[Dict] = None) -> GeneratedProtocolMessageType:
    """Creates a new instance of the given Protobuf class/type and populates it
    with the data given in the dict.

    :param proto_class: Protobuf class to create
    :param dict_data: Python dict with the data to populate the object with
    :return: A populated Protobuf object (message)
    """
    proto = proto_class()
    json_format.ParseDict(dict_data or {}, proto)
    return proto


def _get_proto_field_name(field: dataclasses.Field):
    return field.metadata.get('pb_name', field.name)


def dict_to_dataclass(dataclass_type: Type[T_DCB], dict_data: Optional[Dict] = None, ignore_extras: bool = False) -> T_DCB:
    """Creates a new instance of the given dataclass type and populates it
    (recursively) with the data given in the dict.

    :param dataclass_type: Dataclass type to create
    :param dict_data: Python dict with the data to populate the object with
    :param ignore_extras: Should dict keys that aren't fields in the target
                          dataclass type be ignored or should they raise
                          ValueErrors?
    :return: A populated Dataclass object
    """
    if not dict_data:
        return dataclass_type()
    if isinstance(dict_data, dict):
        new_dict = {}
        field_map = {_get_proto_field_name(f): f for f in dataclasses.fields(dataclass_type)}  # type: Dict[str, dataclasses.Field]
        for k, v in dict_data.items():
            field = field_map.get(k, None)

            if field:
                is_obj = field.metadata.get('is_obj', False)
                is_list = field.metadata.get('is_list', False)
                is_map = field.metadata.get('is_map', False)
                if is_obj:
                    field_type = castutils.get_dc_field_obj_type(field, dataclass_type)
                    if is_map:
                        new_dict[field.name] = {vk: dict_to_dataclass(field_type, vv, ignore_extras) for vk, vv in v.items()}
                    elif is_list:
                        new_dict[field.name] = [dict_to_dataclass(field_type, vv, ignore_extras) for vv in v]
                    else:
                        new_dict[field.name] = dict_to_dataclass(field_type, v, ignore_extras)
                else:
                    dictator = field.metadata.get('dictator', dictators.BaseDictator)
                    if is_map:
                        new_dict[field.name] = {vk: dictator.from_dict_value(vv, field, dataclass_type) for vk, vv in v.items()}
                    elif is_list:
                        new_dict[field.name] = [dictator.from_dict_value(vv, field, dataclass_type) for vv in v]
                    else:
                        new_dict[field.name] = dictator.from_dict_value(v, field, dataclass_type)
            elif not ignore_extras:
                raise ValueError(f'Field "{field.name}" not found in dataclass "{dataclass_type}"')
        return dataclass_type(**new_dict)
    elif isinstance(dict_data, (list, tuple)):
        new_dict = {}
        for field, v in zip(dataclasses.fields(dataclass_type), dict_data):
            dictator = field.metadata.get('dictator', dictators.BaseDictator)
            new_dict[field.name] = dictator.from_dict_value(v, field, dataclass_type)
        return dataclass_type(**new_dict)
    else:
        field = dataclasses.fields(dataclass_type)[0]
        dictator = field.metadata.get('dictator', dictators.BaseDictator)
        return dataclass_type(dictator.from_dict_value(dict_data, field, dataclass_type))



from __future__ import annotations

__all__ = [
    'DataclassBase',
    'T_DCB',  # Required...?!?
]

import json
import base64

from protoplasm.structs._base import *
import dataclasses

import logging
log = logging.getLogger(__name__)

T_DCB = TypeVar('T_DCB', bound='DataclassBase')


@dataclasses.dataclass
class DataclassBase(Generic[T_DCB]):
    __proto_cls__: ClassVar[GeneratedProtocolMessageType]

    def __post_init__(self):
        from protoplasm.casting import castutils
        for f in dataclasses.fields(self):
            if f.metadata.get('is_obj', False):
                f_val = getattr(self, f.name, None)
                if f_val:
                    f_type = castutils.get_dc_field_obj_type(f, self)
                    if f.metadata.get('is_map', False):
                        for k in f_val.keys():
                            f_val[k] = castutils.force_object(f_type, f_val[k])
                    elif f.metadata.get('is_list', False):
                        for i in range(len(f_val)):
                            f_val[i] = castutils.force_object(f_type, f_val[i])
                    else:
                        setattr(self, f.name, castutils.force_object(f_type, f_val))

    def to_dict(self) -> Dict:
        """Turns this `DataclassBase` into a dict of its fields and values
        recursively and returns, so for any field value that is itself a
        `DataclassBase`, `to_dict` is also called, resulting in a nested dict of
        only "primitive" (and JSON serializable) values.
        """
        from protoplasm import casting
        return casting.dataclass_to_dict(self)

    def to_json(self, indent: Optional[int] = None) -> str:
        return json.dumps(self.to_dict(), indent=indent)

    def to_proto(self) -> GeneratedProtocolMessageType:
        from protoplasm import casting
        return casting.dataclass_to_proto(self)

    def to_protostr(self) -> bytes:
        from protoplasm import casting
        return casting.dataclass_to_proto(self).SerializeToString()

    def to_base64(self) -> bytes:
        return base64.encodebytes(self.to_protostr()).strip().replace(b'\r',
                                                                      b'').replace(b'\n', b'').replace(b'\t', b'')

    def unpack(self) -> Dict[str, Any]:
        """Returns a dict with the fields of this `DataclassBase` as keys and
        their values as they are.

        The main difference between this and the `to_dict` method is that
        `to_dict` is recursive so on any field value that is itself a
        `DataclassBase`, `to_dict` is also called, resulting in a nested dict of
        only "primitive" (and JSON serializable) values.

        This `unpack` method on the other hand will leave any `DataclassBase`
        values be.

        The main use-case for this method is to "unpack"/"explode"/"break up" a
        gRPC Proto Service Request/Response message up into individual method
        arguments.
        """
        d = {}
        for field in dataclasses.fields(self):
            d[field.name] = getattr(self, field.name)
        return d

    def explode(self) -> Optional[Union[Tuple[Any], Any]]:
        """Returns a tuple with the field values of this `DataclassBase` or if
        this `DataclassBase` only contains one field, returns that field's value
        alone.

        This is similar to the `unpack()` method except this only returns values
        and if there's only a single field (as many RPC Response messages do),
        it's not wrapped in a tuple.

        The main use-case for this method is to "unpack" a gRPC Proto Service
        RPC Unary Response message up into individual return values (often just
        one).
        """
        tu = []
        for field in dataclasses.fields(self):
            tu.append(getattr(self, field.name))
        if not tu:
            return  # We can theoretically have "empty" Response message classes
        if len(tu) == 1:
            return tu[0]
        return tuple(tu)

    @classmethod
    def from_kwdict(cls, **kwargs) -> T_DCB:
        from protoplasm.casting import castutils
        return cls.from_dict(castutils.kwdict(**kwargs))

    @classmethod
    def from_dict(cls, dict_data: dict, ignore_extras: bool = False) -> T_DCB:
        from protoplasm import casting
        return casting.dict_to_dataclass(cls, dict_data, ignore_extras)

    @classmethod
    def from_json(cls, json_string: str) -> T_DCB:
        return cls.from_dict(json.loads(json_string))

    @classmethod
    def from_proto(cls, proto: GeneratedProtocolMessageType) -> T_DCB:
        from protoplasm import casting
        return cls.from_dict(casting.proto_to_dict(proto))

    @classmethod
    def from_protostr(cls, protostr: bytes) -> T_DCB:
        proto = cls.__proto_cls__()
        proto.ParseFromString(protostr)
        return cls.from_proto(proto)

    @classmethod
    def from_base64(cls, base64str: Union[bytes, str]) -> T_DCB:
        from protoplasm.casting import castutils
        return cls.from_protostr(castutils.fuzzy_base64_to_bytes(base64str))

    @classmethod
    def from_clone(cls, clone: DataclassBase) -> T_DCB:
        return cls.from_dict(clone.to_dict(), ignore_extras=True)



__all__ = [
    'GrpcServer',
]

# TODO(thordurm@ccpgames.com) 2022-06-07: MOVE THIS TO A BETTER LOCATION!!

from concurrent import futures
import grpc  # noqa grpc is in the grpcio package
import time

from protoplasm.bases import grpc_bases

import logging
log = logging.getLogger(__name__)


class GrpcServer(object):
    _ONE_DAY_IN_SECONDS = 60 * 60 * 24

    def __init__(self, port='[::]:50051', max_workers=10):
        """A simple gRPC server.

        Jusd add Servicer implementations (extending
        `protoplasm.bases.grpc_bases.BaseGrpcServicer`) and start it up with
        `serve()`.

        :param port: String listening address and port (default is `[::]:50051`)
        :param max_workers: Maximum workers via `concurrent.futures.ThreadPoolExecutor`
        """
        self.grpc_server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers))
        self.grpc_server.add_insecure_port(port)
        log.info(f'GrpcServer created on port={port}, max_workers={max_workers}')

    def add_servicer(self, servicer: grpc_bases.BaseGrpcServicer):
        """Adds a `protoplasm.bases.grpc_bases.BaseGrpcServicer` based services
        for this server to run.

        :param servicer: A servicer extending `protoplasm.bases.grpc_bases.BaseGrpcServicer`
        """
        servicer.add_to_server(self.grpc_server)
        log.info('Servicer added: %r', servicer)

    def serve(self):
        self.grpc_server.start()
        log.info('Server started')
        try:
            while True:
                log.info('Server listening...')
                time.sleep(GrpcServer._ONE_DAY_IN_SECONDS)
        except KeyboardInterrupt:
            log.info('Server interrupted!')
            self.grpc_server.stop(0)
            log.info('Server Stopped!')



import unittest
import datetime
import base64

from protoplasm.casting import objectifier
import os
import sys

import logging
log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

import shutil
import time
HERE = os.path.dirname(__file__)
PROTO_ROOT = os.path.join(HERE, 'res', 'proto')
BUILD_ROOT = os.path.join(HERE, 'res', 'build')


class ObjectifierTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Remove old stuff...
        build_package = os.path.join(BUILD_ROOT, 'sandbox')
        if os.path.exists(build_package):
            shutil.rmtree(build_package)
            time.sleep(0.1)

        from neobuilder.neobuilder import NeoBuilder

        # Build stuff...
        builder = NeoBuilder(package='sandbox',
                             protopath=PROTO_ROOT,
                             build_root=BUILD_ROOT)
        builder.build()

        # Add build root to path to access its modules
        sys.path.append(BUILD_ROOT)

    def test_dict_to_proto(self):
        from sandbox.test import rainbow_pb2

        p1 = rainbow_pb2.SubMessage()
        p1.foo = 'Foo!'
        p1.bar = 'Bar!!!'

        self.assertEqual(p1, objectifier.dict_to_proto(rainbow_pb2.SubMessage, {'foo': 'Foo!', 'bar': 'Bar!!!'}))

        p2 = rainbow_pb2.SubMessage()
        p2.foo = 'Foo Two!'
        self.assertEqual(p2, objectifier.dict_to_proto(rainbow_pb2.SubMessage, {'foo': 'Foo Two!'}))

        p3 = rainbow_pb2.SubMessage()
        p3.foo = '这是中国人'
        self.assertNotEqual(p3, objectifier.dict_to_proto(rainbow_pb2.SubMessage, {'foo': '这不是中国人'}))

        p4 = rainbow_pb2.SubMessage()
        self.assertEqual(p4, objectifier.dict_to_proto(rainbow_pb2.SubMessage))

    def test_nested_dict_to_proto(self):
        from sandbox.test import rainbow_pb2

        p1 = rainbow_pb2.RainbowMessage()
        self.assertEqual(p1, objectifier.dict_to_proto(rainbow_pb2.RainbowMessage))

        p2 = rainbow_pb2.RainbowMessage()
        p2.simple_field = 'Green'
        self.assertEqual(p2, objectifier.dict_to_proto(rainbow_pb2.RainbowMessage, {'simple_field': 'Green'}))
        self.assertNotEqual(p1, objectifier.dict_to_proto(rainbow_pb2.RainbowMessage, {'simple_field': 'Green'}))

        p3 = rainbow_pb2.RainbowMessage()
        p3.simple_field = 'Green'
        p3.message_field.foo = 'Blue'
        self.assertEqual(p3, objectifier.dict_to_proto(rainbow_pb2.RainbowMessage,
                                                       {'simple_field': 'Green', 'message_field': {'foo': 'Blue'}}))
        p4 = rainbow_pb2.RainbowMessage()
        p4.simple_field = 'Green'
        p4.message_field.foo = 'Blue'
        p4.message_field.bar = 'Yellow'
        self.assertEqual(p4, objectifier.dict_to_proto(rainbow_pb2.RainbowMessage,
                                                       {'simple_field': 'Green',
                                                        'message_field': {'foo': 'Blue', 'bar': 'Yellow'}}))
        self.assertNotEqual(p3, p4)
        self.assertNotEqual(p4, objectifier.dict_to_proto(rainbow_pb2.RainbowMessage,
                                                          {'simple_field': 'Green',
                                                           'message_field': {'foo': 'Blue', 'bar': 'Red'}}))

        p5 = rainbow_pb2.RainbowMessage()
        p5.simple_field = 'Green'
        p5.message_field.foo = 'Blue'
        p5.message_field.bar = 'Yellow'
        p5.simple_list.append('Four')
        p5.simple_list.append('Seven')
        p5.simple_list.append('99')
        self.assertEqual(p5, objectifier.dict_to_proto(rainbow_pb2.RainbowMessage,
                                                       {'simple_field': 'Green', 'simple_list': ['Four', 'Seven', '99'],
                                                        'message_field': {'foo': 'Blue', 'bar': 'Yellow'}}))

        p6 = rainbow_pb2.RainbowMessage()
        p6.simple_field = 'Green'
        p6.message_field.foo = 'Blue'
        p6.message_field.bar = 'Yellow'
        p6.simple_list.append('Four')
        p6.simple_list.append('Seven')
        p6.simple_list.append('98')
        self.assertNotEqual(p6, objectifier.dict_to_proto(rainbow_pb2.RainbowMessage,
                                                          {'simple_field': 'Green',
                                                           'simple_list': ['Four', 'Seven', '100'],
                                                           'message_field': {'foo': 'Blue', 'bar': 'Yellow'}}))

        p7 = rainbow_pb2.RainbowMessage()
        p7.simple_field = 'Green'
        p7.message_field.foo = 'Blue'
        p7.message_field.bar = 'Yellow'
        p7.simple_list.append('Four')
        p7.simple_list.append('Seven')
        p7.simple_list.append('99')
        m1 = p7.message_list.add()
        m1.foo = 'Foo in a list'
        m1.bar = 'Candybar'
        m2 = p7.message_list.add()
        m2.foo = 'a Fool in a list'
        m2.bar = 'Tequila bar'
        self.assertEqual(p7, objectifier.dict_to_proto(rainbow_pb2.RainbowMessage,
                                                       {'simple_field': 'Green', 'simple_list': ['Four', 'Seven', '99'],
                                                        'message_list': [{'foo': 'Foo in a list', 'bar': 'Candybar'},
                                                                         {'foo': 'a Fool in a list',
                                                                          'bar': 'Tequila bar'}],
                                                        'message_field': {'foo': 'Blue', 'bar': 'Yellow'}}))

        self.assertNotEqual(p7, objectifier.dict_to_proto(rainbow_pb2.RainbowMessage,
                                                          {'simple_field': 'Green',
                                                           'simple_list': ['Four', 'Seven', '99'], 'message_list': [
                                                              {'foo': 'a Fool in a list', 'bar': 'Tequila bar'}],
                                                           'message_field': {'foo': 'Blue', 'bar': 'Yellow'}}))

        p8 = rainbow_pb2.RainbowMessage()
        p8.simple_field = 'Green'
        p8.message_field.foo = 'Blue'
        p8.message_field.bar = 'Yellow'
        p8.simple_list.append('Four')
        p8.simple_list.append('Seven')
        p8.simple_list.append('99')
        m3 = p8.message_list.add()
        m3.foo = 'Foo in a list'
        m3.bar = 'Candybar'
        m4 = p8.message_list.add()
        m4.foo = 'a Fool in a list'
        m4.bar = 'Tequila bar'
        p8.simple_map['dora'] = 'Imamap!'
        p8.simple_map['diego'] = 'Camera!'
        self.assertEqual(p8, objectifier.dict_to_proto(rainbow_pb2.RainbowMessage,
                                                       {'simple_field': 'Green', 'simple_list': ['Four', 'Seven', '99'],
                                                        'message_list': [{'foo': 'Foo in a list', 'bar': 'Candybar'},
                                                                         {'foo': 'a Fool in a list',
                                                                          'bar': 'Tequila bar'}],
                                                        'message_field': {'foo': 'Blue', 'bar': 'Yellow'},
                                                        'simple_map': {'dora': 'Imamap!', 'diego': 'Camera!'}}
                                                       ))

        p9 = rainbow_pb2.RainbowMessage()
        p9.simple_field = 'Green'
        p9.message_field.foo = 'Blue'
        p9.message_field.bar = 'Yellow'
        p9.simple_list.append('Four')
        p9.simple_list.append('Seven')
        p9.simple_list.append('99')
        m5 = p9.message_list.add()
        m5.foo = 'Foo in a list'
        m5.bar = 'Candybar'
        m6 = p9.message_list.add()
        m6.foo = 'a Fool in a list'
        m6.bar = 'Tequila bar'
        p9.simple_map['dora'] = 'Imamap!'
        p9.simple_map['diego'] = 'Camera!'
        p9.message_map['donald'].foo = 'duck'
        p9.message_map['donald'].bar = 'trump'
        p9.message_map['mickey'].foo = 'mouse'
        self.assertEqual(p9, objectifier.dict_to_proto(rainbow_pb2.RainbowMessage,
                                                       {'simple_field': 'Green', 'simple_list': ['Four', 'Seven', '99'],
                                                        'message_list': [{'foo': 'Foo in a list', 'bar': 'Candybar'},
                                                                         {'foo': 'a Fool in a list',
                                                                          'bar': 'Tequila bar'}],
                                                        'message_field': {'foo': 'Blue', 'bar': 'Yellow'},
                                                        'simple_map': {'dora': 'Imamap!', 'diego': 'Camera!'},
                                                        'message_map': {'donald': {'foo': 'duck', 'bar': 'trump'},
                                                                        'mickey': {'foo': 'mouse'}}}
                                                       ))

        p10 = rainbow_pb2.RainbowMessage()
        p10.simple_field = 'Green'
        p10.message_field.foo = 'Blue'
        p10.message_field.bar = 'Yellow'
        p10.simple_list.append('Four')
        p10.simple_list.append('Seven')
        p10.simple_list.append('99')
        m7 = p10.message_list.add()
        m7.foo = 'Foo in a list'
        m7.bar = 'Candybar'
        m8 = p10.message_list.add()
        m8.foo = 'a Fool in a list'
        m8.bar = 'Tequila bar'
        p10.simple_map['dora'] = 'Imamap!'
        p10.simple_map['diego'] = 'Camera!'
        p10.message_map['mickey'].foo = 'mouse'
        p10.message_map['donald'].foo = 'duck'
        p10.message_map['donald'].bar = 'trump'

        self.assertEqual(p10, objectifier.dict_to_proto(rainbow_pb2.RainbowMessage,
                                                        {'simple_field': 'Green',
                                                         'simple_list': ['Four', 'Seven', '99'],
                                                         'message_list': [{'foo': 'Foo in a list', 'bar': 'Candybar'},
                                                                          {'foo': 'a Fool in a list',
                                                                           'bar': 'Tequila bar'}],
                                                         'simple_map': {'diego': 'Camera!', 'dora': 'Imamap!'},
                                                         'message_field': {'foo': 'Blue', 'bar': 'Yellow'},
                                                         'message_map': {'donald': {'foo': 'duck', 'bar': 'trump'},
                                                                         'mickey': {'foo': 'mouse'}}}
                                                        ))

    def test_dict_to_dataclass(self):
        from sandbox.test import rainbow_dc

        dc1 = rainbow_dc.SubMessage()
        dc1.foo = 'Foo!'
        dc1.bar = 'Bar!!!'

        self.assertEqual(dc1, objectifier.dict_to_dataclass(rainbow_dc.SubMessage, {'foo': 'Foo!',
                                                                                    'bar': 'Bar!!!'}))

        dc2 = rainbow_dc.SubMessage()
        dc2.foo = 'Foo Two!'
        self.assertEqual(dc2, objectifier.dict_to_dataclass(rainbow_dc.SubMessage, {'foo': 'Foo Two!'}))

        dc3 = rainbow_dc.SubMessage()
        dc3.foo = '这是中国人'
        self.assertNotEqual(dc3, objectifier.dict_to_dataclass(rainbow_dc.SubMessage, {'foo': '这不是中国人'}))

        dc4 = rainbow_dc.SubMessage()
        self.assertEqual(dc4, objectifier.dict_to_dataclass(rainbow_dc.SubMessage))

    def test_nested_dict_to_dataclass(self):
        from sandbox.test import rainbow_dc

        dc1 = rainbow_dc.RainbowMessage()
        self.assertEqual(dc1, objectifier.dict_to_dataclass(rainbow_dc.RainbowMessage))

        dc2 = rainbow_dc.RainbowMessage()
        dc2.simple_field = 'Green'
        self.assertEqual(dc2, objectifier.dict_to_dataclass(rainbow_dc.RainbowMessage, {'simple_field': 'Green'}))
        self.assertNotEqual(dc1, objectifier.dict_to_dataclass(rainbow_dc.RainbowMessage, {'simple_field': 'Green'}))

        dc3 = rainbow_dc.RainbowMessage()
        dc3.simple_field = 'Green'
        dc3.message_field = rainbow_dc.SubMessage()
        dc3.message_field.foo = 'Blue'
        self.assertEqual(dc3, objectifier.dict_to_dataclass(rainbow_dc.RainbowMessage,
                                                       {'simple_field': 'Green', 'message_field': {'foo': 'Blue'}}))
        dc4 = rainbow_dc.RainbowMessage()
        dc4.simple_field = 'Green'
        dc4.message_field = rainbow_dc.SubMessage()
        dc4.message_field.foo = 'Blue'
        dc4.message_field.bar = 'Yellow'
        self.assertEqual(dc4, objectifier.dict_to_dataclass(rainbow_dc.RainbowMessage,
                                                       {'simple_field': 'Green',
                                                        'message_field': {'foo': 'Blue', 'bar': 'Yellow'}}))
        self.assertNotEqual(dc3, dc4)
        self.assertNotEqual(dc4, objectifier.dict_to_dataclass(rainbow_dc.RainbowMessage,
                                                          {'simple_field': 'Green',
                                                           'message_field': {'foo': 'Blue', 'bar': 'Red'}}))

        dc5 = rainbow_dc.RainbowMessage()
        dc5.simple_field = 'Green'
        dc5.message_field = rainbow_dc.SubMessage('Blue', 'Yellow')
        dc5.simple_list.append('Four')
        dc5.simple_list.append('Seven')
        dc5.simple_list.append('99')
        self.assertEqual(dc5, objectifier.dict_to_dataclass(rainbow_dc.RainbowMessage,
                                                       {'simple_field': 'Green', 'simple_list': ['Four', 'Seven', '99'],
                                                        'message_field': {'foo': 'Blue', 'bar': 'Yellow'}}))

        dc6 = rainbow_dc.RainbowMessage()
        dc6.simple_field = 'Green'
        dc6.message_field = rainbow_dc.SubMessage(foo='Blue', bar='Yellow')
        dc6.simple_list.append('Four')
        dc6.simple_list.append('Seven')
        dc6.simple_list.append('98')
        self.assertNotEqual(dc6, objectifier.dict_to_dataclass(rainbow_dc.RainbowMessage,
                                                          {'simple_field': 'Green',
                                                           'simple_list': ['Four', 'Seven', '100'],
                                                           'message_field': {'foo': 'Blue', 'bar': 'Yellow'}}))

        dc7 = rainbow_dc.RainbowMessage()
        dc7.simple_field = 'Green'
        dc7.message_field = rainbow_dc.SubMessage(foo='Blue', bar='Yellow')
        dc7.simple_list.append('Four')
        dc7.simple_list.append('Seven')
        dc7.simple_list.append('99')

        dc7.message_list.append(rainbow_dc.SubMessage(foo='Foo in a list', bar='Candybar'))
        dc7.message_list.append(rainbow_dc.SubMessage(foo='a Fool in a list', bar='Tequila bar'))

        self.assertEqual(dc7, objectifier.dict_to_dataclass(rainbow_dc.RainbowMessage,
                                                       {'simple_field': 'Green', 'simple_list': ['Four', 'Seven', '99'],
                                                        'message_list': [{'foo': 'Foo in a list', 'bar': 'Candybar'},
                                                                         {'foo': 'a Fool in a list',
                                                                          'bar': 'Tequila bar'}],
                                                        'message_field': {'foo': 'Blue', 'bar': 'Yellow'}}))

        self.assertNotEqual(dc7, objectifier.dict_to_dataclass(rainbow_dc.RainbowMessage,
                                                          {'simple_field': 'Green',
                                                           'simple_list': ['Four', 'Seven', '99'], 'message_list': [
                                                              {'foo': 'a Fool in a list', 'bar': 'Tequila bar'}],
                                                           'message_field': {'foo': 'Blue', 'bar': 'Yellow'}}))

        dc8 = rainbow_dc.RainbowMessage()
        dc8.simple_field = 'Green'
        dc8.message_field = rainbow_dc.SubMessage(foo='Blue', bar='Yellow')
        dc8.simple_list.append('Four')
        dc8.simple_list.append('Seven')
        dc8.simple_list.append('99')

        dc8.message_list.append(rainbow_dc.SubMessage(foo='Foo in a list', bar='Candybar'))
        dc8.message_list.append(rainbow_dc.SubMessage(foo='a Fool in a list', bar='Tequila bar'))

        dc8.simple_map['dora'] = 'Imamap!'
        dc8.simple_map['diego'] = 'Camera!'
        self.assertEqual(dc8, objectifier.dict_to_dataclass(rainbow_dc.RainbowMessage,
                                                       {'simple_field': 'Green', 'simple_list': ['Four', 'Seven', '99'],
                                                        'message_list': [{'foo': 'Foo in a list', 'bar': 'Candybar'},
                                                                         {'foo': 'a Fool in a list',
                                                                          'bar': 'Tequila bar'}],
                                                        'message_field': {'foo': 'Blue', 'bar': 'Yellow'},
                                                        'simple_map': {'dora': 'Imamap!', 'diego': 'Camera!'}}
                                                       ))

        dc9 = rainbow_dc.RainbowMessage()
        dc9.simple_field = 'Green'
        dc9.message_field = rainbow_dc.SubMessage(foo='Blue', bar='Yellow')
        dc9.simple_list.append('Four')
        dc9.simple_list.append('Seven')
        dc9.simple_list.append('99')
        dc9.message_list.append(rainbow_dc.SubMessage(foo='Foo in a list', bar='Candybar'))
        dc9.message_list.append(rainbow_dc.SubMessage(foo='a Fool in a list', bar='Tequila bar'))
        dc9.simple_map['dora'] = 'Imamap!'
        dc9.simple_map['diego'] = 'Camera!'

        dc9.message_map['donald'] = rainbow_dc.SubMessage(foo='duck', bar='trump')

        dc9.message_map['mickey'] = rainbow_dc.SubMessage(foo='mouse')

        self.assertEqual(dc9, objectifier.dict_to_dataclass(rainbow_dc.RainbowMessage,
                                                       {'simple_field': 'Green', 'simple_list': ['Four', 'Seven', '99'],
                                                        'message_list': [{'foo': 'Foo in a list', 'bar': 'Candybar'},
                                                                         {'foo': 'a Fool in a list',
                                                                          'bar': 'Tequila bar'}],
                                                        'message_field': {'foo': 'Blue', 'bar': 'Yellow'},
                                                        'simple_map': {'dora': 'Imamap!', 'diego': 'Camera!'},
                                                        'message_map': {'donald': {'foo': 'duck', 'bar': 'trump'},
                                                                        'mickey': {'foo': 'mouse'}}}
                                                       ))

        dc10 = rainbow_dc.RainbowMessage()
        dc10.simple_field = 'Green'
        dc10.message_field = rainbow_dc.SubMessage(foo='Blue', bar='Yellow')
        dc10.simple_list.append('Four')
        dc10.simple_list.append('Seven')
        dc10.simple_list.append('99')
        dc10.message_list.append(rainbow_dc.SubMessage(foo='Foo in a list', bar='Candybar'))
        dc10.message_list.append(rainbow_dc.SubMessage(foo='a Fool in a list', bar='Tequila bar'))
        dc10.simple_map['dora'] = 'Imamap!'
        dc10.simple_map['diego'] = 'Camera!'

        dc10.message_map['mickey'] = rainbow_dc.SubMessage(foo='mouse')
        dc10.message_map['donald'] = rainbow_dc.SubMessage(foo='duck', bar='trump')

        self.assertEqual(dc10, objectifier.dict_to_dataclass(rainbow_dc.RainbowMessage,
                                                        {'simple_field': 'Green',
                                                         'simple_list': ['Four', 'Seven', '99'],
                                                         'message_list': [{'foo': 'Foo in a list', 'bar': 'Candybar'},
                                                                          {'foo': 'a Fool in a list',
                                                                           'bar': 'Tequila bar'}],
                                                         'simple_map': {'diego': 'Camera!', 'dora': 'Imamap!'},
                                                         'message_field': {'foo': 'Blue', 'bar': 'Yellow'},
                                                         'message_map': {'donald': {'foo': 'duck', 'bar': 'trump'},
                                                                         'mickey': {'foo': 'mouse'}}}
                                                        ))

    def test_timestamp_dict_to_proto(self):
        from sandbox.test import rainbow_pb2

        ts1 = '2012-07-03T14:50:51.654321Z'
        dt1 = datetime.datetime(2012, 7, 3, 14, 50, 51, 654321)
        self.assertEqual(ts1, dt1.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts2 = '2010-04-09T22:38:39.009870Z'
        dt2 = datetime.datetime(2010, 4, 9, 22, 38, 39, 9870)
        self.assertEqual(ts2, dt2.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts3 = '1979-07-06T14:30:00.000000Z'
        dt3 = datetime.datetime(1979, 7, 6, 14, 30)
        self.assertEqual(ts3, dt3.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts4 = '2049-01-01T12:00:00.000000Z'
        dt4 = datetime.datetime(2049, 1, 1, 12)
        self.assertEqual(ts4, dt4.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts5 = '2049-01-01T00:00:00.000000Z'
        dt5 = datetime.datetime(2049, 1, 1)
        self.assertEqual(ts5, dt5.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        dict_data = {'my_timestamp': ts1,
                     'my_timestamp_list': [ts2, ts3],
                     'my_timestamp_map': {'noon': ts4,
                                          'midnight': ts5}}

        p_expect = rainbow_pb2.TimestampMessage()
        p_expect.my_timestamp.FromJsonString(ts1)
        p_expect.my_timestamp_list.add().FromJsonString(ts2)
        p_expect.my_timestamp_list.add().FromJsonString(ts3)
        p_expect.my_timestamp_map['noon'].FromJsonString(ts4)
        p_expect.my_timestamp_map['midnight'].FromJsonString(ts5)

        self.assertEqual(p_expect, objectifier.dict_to_proto(rainbow_pb2.TimestampMessage, dict_data))

    def test_timestamp_dict_to_dataclass(self):
        from sandbox.test import rainbow_dc

        ts1 = '2012-07-03T14:50:51.654321Z'
        dt1 = datetime.datetime(2012, 7, 3, 14, 50, 51, 654321)
        self.assertEqual(ts1, dt1.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts2 = '2010-04-09T22:38:39.009870Z'
        dt2 = datetime.datetime(2010, 4, 9, 22, 38, 39, 9870)
        self.assertEqual(ts2, dt2.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts3 = '1979-07-06T14:30:00.000000Z'
        dt3 = datetime.datetime(1979, 7, 6, 14, 30)
        self.assertEqual(ts3, dt3.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts4 = '2049-01-01T12:00:00.000000Z'
        dt4 = datetime.datetime(2049, 1, 1, 12)
        self.assertEqual(ts4, dt4.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts5 = '2049-01-01T00:00:00.000000Z'
        dt5 = datetime.datetime(2049, 1, 1)
        self.assertEqual(ts5, dt5.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        dict_data = {'my_timestamp': ts1,
                     'my_timestamp_list': [ts2, ts3],
                     'my_timestamp_map': {'noon': ts4,
                                          'midnight': ts5}}

        dc_expect = rainbow_dc.TimestampMessage()
        dc_expect.my_timestamp = dt1
        dc_expect.my_timestamp_list.append(dt2)
        dc_expect.my_timestamp_list.append(dt3)
        dc_expect.my_timestamp_map['noon'] = dt4
        dc_expect.my_timestamp_map['midnight'] = dt5

        self.assertEqual(dc_expect, objectifier.dict_to_dataclass(rainbow_dc.TimestampMessage, dict_data))

    def test_byte_dict_to_proto(self):
        from sandbox.test import rainbow_pb2

        as_str_1 = 'Þórður Matthíasson'
        as_bytes_1 = as_str_1.encode('utf-8')
        as_base64_1 = base64.encodebytes(as_bytes_1).decode('utf-8').strip()

        as_str_2 = '♡'
        as_bytes_2 = as_str_2.encode('utf-8')
        as_base64_2 = base64.encodebytes(as_bytes_2).decode('utf-8').strip()

        as_str_3 = '♠'
        as_bytes_3 = as_str_3.encode('utf-8')
        as_base64_3 = base64.encodebytes(as_bytes_3).decode('utf-8').strip()

        as_str_4 = '♢'
        as_bytes_4 = as_str_4.encode('utf-8')
        as_base64_4 = base64.encodebytes(as_bytes_4).decode('utf-8').strip()

        as_str_5 = '♣'
        as_bytes_5 = as_str_5.encode('utf-8')
        as_base64_5 = base64.encodebytes(as_bytes_5).decode('utf-8').strip()

        dict_data = {'my_bytes': as_base64_1,
                     'my_bytes_list': [as_base64_2, as_base64_3],
                     'my_bytes_map': {'zero': as_base64_4,
                                      'one': as_base64_5}}

        p_expect = rainbow_pb2.BytesMessage()
        p_expect.my_bytes = as_bytes_1
        p_expect.my_bytes_list.append(as_bytes_2)
        p_expect.my_bytes_list.append(as_bytes_3)
        p_expect.my_bytes_map['zero'] = as_bytes_4
        p_expect.my_bytes_map['one'] = as_bytes_5

        self.assertEqual(p_expect, objectifier.dict_to_proto(rainbow_pb2.BytesMessage, dict_data))

    def test_byte_dict_to_dataclass(self):
        from sandbox.test import rainbow_dc

        as_str_1 = 'Þórður Matthíasson'
        as_bytes_1 = as_str_1.encode('utf-8')
        as_base64_1 = base64.encodebytes(as_bytes_1).decode('utf-8').strip()

        as_str_2 = '♡'
        as_bytes_2 = as_str_2.encode('utf-8')
        as_base64_2 = base64.encodebytes(as_bytes_2).decode('utf-8').strip()

        as_str_3 = '♠'
        as_bytes_3 = as_str_3.encode('utf-8')
        as_base64_3 = base64.encodebytes(as_bytes_3).decode('utf-8').strip()

        as_str_4 = '♢'
        as_bytes_4 = as_str_4.encode('utf-8')
        as_base64_4 = base64.encodebytes(as_bytes_4).decode('utf-8').strip()

        as_str_5 = '♣'
        as_bytes_5 = as_str_5.encode('utf-8')
        as_base64_5 = base64.encodebytes(as_bytes_5).decode('utf-8').strip()

        dict_data = {'my_bytes': as_base64_1,
                     'my_bytes_list': [as_base64_2, as_base64_3],
                     'my_bytes_map': {'zero': as_base64_4,
                                      'one': as_base64_5}}

        dc_expect = rainbow_dc.BytesMessage()
        dc_expect.my_bytes = as_bytes_1
        dc_expect.my_bytes_list.append(as_bytes_2)
        dc_expect.my_bytes_list.append(as_bytes_3)
        dc_expect.my_bytes_map['zero'] = as_bytes_4
        dc_expect.my_bytes_map['one'] = as_bytes_5

        self.assertEqual(dc_expect, objectifier.dict_to_dataclass(rainbow_dc.BytesMessage, dict_data))

    def test_enum_dict_to_proto(self):
        from sandbox.test import enums_pb2

        p_expect = enums_pb2.WithExternalEnum()
        p_expect.my_enum = enums_pb2.TWO

        p_expect.my_enum_list.append(enums_pb2.ONE)
        p_expect.my_enum_list.append(enums_pb2.THREE)

        p_expect.my_enum_map['one'] = enums_pb2.ONE
        p_expect.my_enum_map['two'] = enums_pb2.TWO

        p_expect.my_alias_enum = enums_pb2.SIX

        p_expect.my_alias_enum_list.append(enums_pb2.FOUR)
        p_expect.my_alias_enum_list.append(enums_pb2.FIVE)
        p_expect.my_alias_enum_list.append(enums_pb2.FJORIR)

        p_expect.my_alias_enum_map['six'] = enums_pb2.SIX
        p_expect.my_alias_enum_map['sex'] = enums_pb2.SEX
        p_expect.my_alias_enum_map['fimm'] = enums_pb2.FIVE

        dict_data = {'my_enum': 2,
                    'my_alias_enum': 3,
                    'my_enum_list': [1, 3],
                    'my_alias_enum_list': [1, 2, 1],
                    'my_enum_map': {'one': 1,
                                    'two': 2},
                    'my_alias_enum_map': {'six': 3,
                                          'sex': 3,
                                          'fimm': 2}}

        self.assertEqual(p_expect, objectifier.dict_to_proto(enums_pb2.WithExternalEnum, dict_data))

        p_expect2 = enums_pb2.WithInternalEnum()
        p_expect2.my_internal_enum = enums_pb2.WithInternalEnum.SIX

        p_expect2.my_internal_enum_list.append(enums_pb2.WithInternalEnum.FIVE)
        p_expect2.my_internal_enum_list.append(enums_pb2.WithInternalEnum.FOUR)

        p_expect2.my_internal_enum_map['no5'] = enums_pb2.WithInternalEnum.FIVE
        p_expect2.my_internal_enum_map['no6'] = enums_pb2.WithInternalEnum.SIX

        p_expect2.my_internal_alias_enum = enums_pb2.WithInternalEnum.SEVEN

        p_expect2.my_internal_alias_enum_list.append(enums_pb2.WithInternalEnum.SJO)
        p_expect2.my_internal_alias_enum_list.append(enums_pb2.WithInternalEnum.ATTA)
        p_expect2.my_internal_alias_enum_list.append(enums_pb2.WithInternalEnum.EIGHT)

        p_expect2.my_internal_alias_enum_map['no9'] = enums_pb2.WithInternalEnum.NIU
        p_expect2.my_internal_alias_enum_map['no9B'] = enums_pb2.WithInternalEnum.NINE
        p_expect2.my_internal_alias_enum_map['default'] = enums_pb2.WithInternalEnum.ZERO

        dict_data2 = {'my_internal_enum': 6,
                      'my_internal_alias_enum': 7,
                      'my_internal_enum_list': [5, 4],
                      'my_internal_alias_enum_list': [7, 8, 8],
                      'my_internal_enum_map': {'no5': 5,
                                               'no6': 6},
                      'my_internal_alias_enum_map': {'no9': 9,
                                                     'no9B': 9,
                                                     'default': 0}}

        self.assertEqual(p_expect2, objectifier.dict_to_proto(enums_pb2.WithInternalEnum, dict_data2))

    def test_enum_dict_to_dataclass(self):
        from sandbox.test import enums_dc

        p_expect = enums_dc.WithExternalEnum()
        p_expect.my_enum = enums_dc.TWO

        p_expect.my_enum_list.append(enums_dc.ONE)
        p_expect.my_enum_list.append(enums_dc.THREE)

        p_expect.my_enum_map['one'] = enums_dc.ONE
        p_expect.my_enum_map['two'] = enums_dc.TWO

        p_expect.my_alias_enum = enums_dc.SIX

        p_expect.my_alias_enum_list.append(enums_dc.FOUR)
        p_expect.my_alias_enum_list.append(enums_dc.FIVE)
        p_expect.my_alias_enum_list.append(enums_dc.FJORIR)

        p_expect.my_alias_enum_map['six'] = enums_dc.SIX
        p_expect.my_alias_enum_map['sex'] = enums_dc.SEX
        p_expect.my_alias_enum_map['fimm'] = enums_dc.FIVE

        dict_data = {'my_enum': 2,
                    'my_alias_enum': 3,
                    'my_enum_list': [1, 3],
                    'my_alias_enum_list': [1, 2, 1],
                    'my_enum_map': {'one': 1,
                                    'two': 2},
                    'my_alias_enum_map': {'six': 3,
                                          'sex': 3,
                                          'fimm': 2}}

        self.assertEqual(p_expect, objectifier.dict_to_dataclass(enums_dc.WithExternalEnum, dict_data))

        p_expect2 = enums_dc.WithInternalEnum()
        p_expect2.my_internal_enum = enums_dc.WithInternalEnum.SIX

        p_expect2.my_internal_enum_list.append(enums_dc.WithInternalEnum.FIVE)
        p_expect2.my_internal_enum_list.append(enums_dc.WithInternalEnum.FOUR)

        p_expect2.my_internal_enum_map['no5'] = enums_dc.WithInternalEnum.FIVE
        p_expect2.my_internal_enum_map['no6'] = enums_dc.WithInternalEnum.SIX

        p_expect2.my_internal_alias_enum = enums_dc.WithInternalEnum.SEVEN

        p_expect2.my_internal_alias_enum_list.append(enums_dc.WithInternalEnum.SJO)
        p_expect2.my_internal_alias_enum_list.append(enums_dc.WithInternalEnum.ATTA)
        p_expect2.my_internal_alias_enum_list.append(enums_dc.WithInternalEnum.EIGHT)

        p_expect2.my_internal_alias_enum_map['no9'] = enums_dc.WithInternalEnum.NIU
        p_expect2.my_internal_alias_enum_map['no9B'] = enums_dc.WithInternalEnum.NINE
        p_expect2.my_internal_alias_enum_map['default'] = enums_dc.WithInternalEnum.ZERO

        dict_data2 = {'my_internal_enum': 6,
                      'my_internal_alias_enum': 7,
                      'my_internal_enum_list': [5, 4],
                      'my_internal_alias_enum_list': [7, 8, 8],
                      'my_internal_enum_map': {'no5': 5,
                                               'no6': 6},
                      'my_internal_alias_enum_map': {'no9': 9,
                                                     'no9B': 9,
                                                     'default': 0}}

        self.assertEqual(p_expect2, objectifier.dict_to_dataclass(enums_dc.WithInternalEnum, dict_data2))

    def test_struct_dict_to_proto(self):
        from sandbox.test import googlestruct_pb2

        the_dict = {
            'my_struct': {
                'my_bool': True,
                'my_float': 4.2,
                'my_null': None,
                'my_dict': {'foo': 'bar', 'you': 'tube'},
                'my_list': [1.0, 3.0, 5.0, 8.0],
                'my_string': 'I am String, hear me spell!',
                'my_int': 42.0
            }
        }

        expected_pb = googlestruct_pb2.StructMessage()
        expected_pb.my_struct['my_string'] = 'I am String, hear me spell!'
        expected_pb.my_struct['my_int'] = 42
        expected_pb.my_struct['my_float'] = 4.2
        expected_pb.my_struct['my_null'] = None
        expected_pb.my_struct['my_bool'] = True
        expected_pb.my_struct['my_list'] = [1, 3, 5, 8]
        expected_pb.my_struct['my_dict'] = {'foo': 'bar', 'you': 'tube'}

        self.assertEqual(expected_pb, objectifier.dict_to_proto(googlestruct_pb2.StructMessage, the_dict))

    def test_struct_dict_to_dataclass(self):
        from sandbox.test.googlestruct_dc import StructMessage

        the_dict = {
            'my_struct': {
                'my_bool': True,
                'my_float': 4.2,
                'my_null': None,
                'my_dict': {'foo': 'bar', 'you': 'tube'},
                'my_list': [1.0, 3.0, 5.0, 8.0],
                'my_string': 'I am String, hear me spell!',
                'my_int': 42.0
            }
        }

        struct_dc = StructMessage(my_struct={
            'my_bool': True,
            'my_float': 4.2,
            'my_null': None,
            'my_dict': {'foo': 'bar', 'you': 'tube'},
            'my_list': [1.0, 3.0, 5.0, 8.0],
            'my_string': 'I am String, hear me spell!',
            'my_int': 42.0
        })

        self.assertEqual(struct_dc, objectifier.dict_to_dataclass(StructMessage, the_dict))



__all__ = [
    'get_dataclass_by_message_name',
]

import typing
from google.protobuf import symbol_database
from protoplasm.bases import dataclass_bases
from protoplasm.casting import castutils

import logging
log = logging.getLogger(__name__)


def get_dataclass_by_message_name(message_name: str) -> typing.Type[dataclass_bases.DataclassBase]:
    sym_db = symbol_database.Default()
    proto_msg = sym_db.pool.FindMessageTypeByName(message_name)
    return castutils.import_dataclass_by_proto(proto_msg._concrete_class)
